__all__ = [
    'VulnerabilityFile',
    'EventsFile',
    'FootprintBinFile',
    'FootprintIdxFile',
    'DamageBinDictFile',
    'OccurrenceFile',
    'RandomFile',
    'CoveragesFile',
    'ItemsFile',
    'FMProgrammeFile',
    'FMPolicyTCFile',
    'FMProfileFile',
    'FMXrefFile',
    'GULSummaryXrefFile',
    'FMSummaryXrefFile'
]

import os
import struct
from collections import OrderedDict
from math import erf

import numpy as np


class ModelFile:
    """
    Base class for all dummy model files.

    Each dummy model file is a class that inherits from this base class. The
    typical order of execution is as follows:
        1. Initialise class attributes and methods (__init__).
        2. Set random seed (seed_rng).
        3. Generate random data (generate_data).
        4. Convert random data to binary format and write to file. This step is
        done as each line of data is generated to minimise memory use
        (write_file).

    Attributes:
        seed_rng: Seed random number generator.
        write_file: Write data to output file in binary format.
        debug_write_file: Write data to screen in csv format.
        generate_data: Generate dummy model data.
    """

    def __init__(self):
        pass

    def seed_rng(self):
        """
        Seed random number generator.

        Assign different random number generator seed to generate each
        randomised dummy model file data. Pollute seeds with salt to prevent
        all random number generators starting with the same seed.
        """
        if self.random_seed == 0:
            np.random.seed()
        elif self.random_seed == -1:
            # Add salt to random seed using name of child class
            salt = int.from_bytes(type(self).__name__.encode(), 'little')
            np.random.seed((1234 + salt) % 0xFFFFFFFF)
        else:
            # Add salt to random seed using name of child class
            salt = int.from_bytes(type(self).__name__.encode(), 'little')
            np.random.seed((self.random_seed + salt) % 0xFFFFFFFF)

    def write_file(self):
        """
        Write data to output file in binary format.

        General method to convert generated data to binary format and write to
        file. Calls chlid class-specific generate_data method.
        """
        with open(self.file_name, 'wb') as f:
            if self.start_stats:
                for stat in self.start_stats:
                    f.write(struct.pack(stat['dtype'], stat['value']))
            dtypes_list = ''.join(self.dtypes.values())
            for line in self.generate_data():
                f.write(struct.pack('=' + dtypes_list, *(line)))

    def debug_write_file(self):
        """
        Write data to screen in csv format.

        Used for debugging file output.
        """
        if self.start_stats:
            for stat in self.start_stats:
                print('{} = {}'.format(stat['desc'], stat['value']))
        line_format = '{}' + ',{}' * (len(self.dtypes) - 1)
        print(line_format.format(*self.dtypes.keys()))
        for line in self.generate_data():
            print(line_format.format(*line))

    def generate_data(self):
        """
        Generate dummy model data.

        Class specific method to generate randomised data. Is called by
        write_file method.
        """
        pass


class VulnerabilityFile(ModelFile):
    """
    Generate random data for Vulnerability dummy model file.

    This file shows the conditional distributions of damage for each intensity
    bin and for each vulnerability ID.

    Attributes:
        generate_data: Generate Vulnerability dummy model file data.
    """

    def __init__(
        self, num_vulnerabilities, num_intensity_bins, num_damage_bins,
        vulnerability_sparseness, random_seed, directory
    ):
        """
        Initialise VulnerabilityFile class.

        Args:
            num_vulnerabilities (int): number of vulnerabilities.
            num_intensity_bins (int): number of intensity bins.
            num_damage_bins(int): number of damage bins.
            vulnerability_sparseness (float): percentage of bins normalised to
                range [0,1] impacted for a vulnerability at an intensity level.
            random_seed (float): random seed for random number generator.
            directory (str): dummy model file destination.
        """
        self.num_vulnerabilities = num_vulnerabilities
        self.num_intensity_bins = num_intensity_bins
        self.num_damage_bins = num_damage_bins
        self.vulnerability_sparseness = vulnerability_sparseness
        self.dtypes = OrderedDict([
            ('vulnerability_id', 'i'), ('intensity_bin_index', 'i'),
            ('damage_bin_index', 'i'), ('prob', 'f')
        ])
        self.start_stats = [
            {
                'desc': 'Number of damage bins', 'value': num_damage_bins,
                'dtype': 'i'
            }
        ]
        self.random_seed = random_seed
        self.data_length = num_vulnerabilities * num_intensity_bins * num_damage_bins
        self.file_name = os.path.join(directory, 'vulnerability.bin')

    def generate_data(self):
        """
        Generate Vulnerability dummy model file data.

        Yields:
            vulnerability (int): vulnerability ID.
            intensity_bin (int): intensity bin ID.
            damage_bin (int): damage bin ID.
            probability (float): impact probability.
        """
        super().seed_rng()
        for vulnerability in range(self.num_vulnerabilities):
            for intensity_bin in range(self.num_intensity_bins):

                # Generate probabalities according to vulnerability sparseness
                # and normalise
                triggers = np.random.uniform(size=self.num_damage_bins)
                probabilities = np.apply_along_axis(
                    lambda x: np.where(
                        x < self.vulnerability_sparseness,
                        np.random.uniform(size=x.shape), 0.0
                    ), 0, triggers
                )
                total_probability = np.sum(probabilities)
                if (total_probability == 0):
                    probabilities[0] = 1.0   # First damage bin is always zero-loss
                else:
                    probabilities /= total_probability

                for damage_bin, probability in enumerate(probabilities):
                    yield vulnerability + 1, intensity_bin + 1, damage_bin + 1, probability


class EventsFile(ModelFile):
    """
    Generate random data for Events dummy model file.

    This file lists event IDs to be run.

    Attributes:
        generate_data: Generate Events dummy model file data.
    """

    def __init__(self, num_events, directory):
        """
        Initialise VulnerabilityFile class.

        Args:
            num_events (int): number of events.
            directory (str): dummy model file destination.
        """
        self.num_events = num_events
        self.dtypes = {'event_id': 'i'}
        self.start_stats = None
        self.data_length = num_events
        self.file_name = os.path.join(directory, 'events.bin')

    def generate_data(self):
        """
        Generate Events dummy model file data.

        Yields:
            event (int): event ID.
        """
        return (tuple([event]) for event in range(1, self.num_events + 1))


class FootprintIdxFile(ModelFile):
    """
    Generate data for Footprint index dummy model file.

    The binary footprint file footprint.bin requires the index file
    footprint.idx.

    Attributes:
        write_file: Write data to Footprint index file in binary format.
    """

    def __init__(self, directory):
        """
        Initialise Footprint index file class.

        Args:
            directory (str): dummy model file destination.
        """
        self.dtypes = OrderedDict([
            ('event_id', 'i'), ('offset', 'q'), ('size', 'q')
        ])
        self.dtypes_list = ''.join(self.dtypes.values())
        self.file_name = os.path.join(directory, 'footprint.idx')

    def write_file(self, event_id, offset, event_size):
        """
        Write data to output Footprint index file in binary format.

        Overrides method in base class. Converts data to arguments to binary and
        writes to file. Called by FootprintBinFile.generate_data().

        Args:
            event_id (int): event ID.
            offset (long long): position of data for event ID in generated
                Footprint binary file relative to beginning of that file.
            size (long long): size of data corresponding to event ID in
                generated Footprint binary file, long long.
        """
        with open(self.file_name, 'ab') as f:
            f.write(struct.pack(
                '=' + self.dtypes_list, event_id, offset, event_size)
            )


class FootprintBinFile(ModelFile):
    """
    Generate data for Footprint binary dummy model file.

    This file shows the intensity of a given event-areaperil combination. The
    binary footprint file footprint.bin requires the index file footprint.idx.

    Attributes:
        generate_data: Generate Footprint binary dummy model file data.
    """

    def __init__(
        self, num_events, num_areaperils, areaperils_per_event,
        num_intensity_bins, intensity_sparseness, no_intensity_uncertainty,
        random_seed, directory
    ):
        """
        Initialise Footprint binary file class.

        Args:
            num_events (int): number of events.
            num_areaperils (int): number of areaperils.
            areaperils_per_event (int): number of areaperils impacted per event.
            num_intensity_bins (int): number of intensity bins.
            intensity_sparseness (float): percentage of bins normalised to
                range [0,1] impacted for an event and areaperil.
            no_intensity_uncertainty (bool): flag to indicate whether more than
                one intensity bin can be impacted, bool.
            random_seed (float): random seed for random number generator.
            directory (str): dummy model file destination.
        """
        self.num_events = num_events
        self.num_areaperils = num_areaperils
        self.areaperils_per_event = areaperils_per_event
        self.num_intensity_bins = num_intensity_bins
        self.intensity_sparseness = intensity_sparseness
        self.no_intensity_uncertainty = no_intensity_uncertainty
        self.random_seed = random_seed
        self.file_name = os.path.join(directory, 'footprint.bin')
        self.event_id = 0
        self.start_stats = [
            {
                'desc': 'Number of intensity bins',
                'value': self.num_intensity_bins, 'dtype': 'i'
            },
            {
                'desc': 'Has Intensity Uncertainty',
                'value': not self.no_intensity_uncertainty, 'dtype': 'i'
            }
        ]
        self.dtypes = OrderedDict(
            [
                ('areaperil_id', 'i'),
                ('intensity_bin_id', 'i'),
                ('probability', 'f')
            ]
        )
        self.idx_file = FootprintIdxFile(directory)
        # Size of data is the same for all events
        self.size = 0
        for dtype in self.dtypes.values():
            self.size += struct.calcsize(dtype)
        if not self.no_intensity_uncertainty:
            self.size *= self.num_intensity_bins
        # Set initial offset
        self.offset = 0
        for stat in self.start_stats:
            self.offset += struct.calcsize(stat['dtype'])

    def generate_data(self):
        """
        Generate Footprint binary dummy model file data.

        Yields:
            areaperil (int): areaperil ID.
            intensity_bin (int): intensity bin ID.
            probability (float): impact probability.
        """
        super().seed_rng()
        for event in range(self.num_events):
            event_size = 0

            if self.areaperils_per_event == self.num_areaperils:
                selected_areaperils = np.arange(1, self.num_areaperils + 1)
            else:
                selected_areaperils = np.random.choice(
                    self.num_areaperils, self.areaperils_per_event,
                    replace=False
                )
                selected_areaperils += 1
                selected_areaperils = np.sort(selected_areaperils)

            for areaperil in selected_areaperils:
                if self.no_intensity_uncertainty:
                    intensity_bin = np.random.randint(
                        1, self.num_intensity_bins + 1
                    )
                    probability = 1.0
                    event_size += self.size
                    yield areaperil, intensity_bin, probability
                else:
                    # Generate probabalities according to intensity sparseness
                    # and normalise
                    triggers = np.random.uniform(size=self.num_intensity_bins)
                    probabilities = np.apply_along_axis(
                        lambda x: np.where(
                            x < self.intensity_sparseness,
                            np.random.uniform(size=x.shape), 0.0
                        ), 0, triggers
                    )
                    total_probability = np.sum(probabilities)
                    if total_probability == 0:
                        continue   # No impacted intensity bins
                    event_size += self.size
                    probabilities /= total_probability

                    for intensity_bin, probability in enumerate(probabilities):
                        yield areaperil, intensity_bin + 1, probability

            self.idx_file.write_file(event + 1, self.offset, event_size)
            self.offset += event_size


class DamageBinDictFile(ModelFile):
    """
    Generate data for Damage Bin Dictionary dummy model file.

    This file shows the discretisation of the effective damageability cumulative
    distribution function.

    Attributes:
        generate_data: Generate Damage Bin Dictionary dummy model file data.
    """

    def __init__(self, num_damage_bins, directory):
        """
        Initialise Damage Bin Dictionary file class.

        Args:
            num_damage_bins (int): number of damage bins.
            directory (str): dummy model file destination.
        """
        self.num_damage_bins = num_damage_bins
        self.dtypes = OrderedDict([
            ('bin_index', 'i'), ('bin_from', 'f'), ('bin_to', 'f'),
            ('interpolation', 'f'), ('interval_type', 'i')
        ])
        self.start_stats = None
        self.data_length = num_damage_bins
        self.file_name = os.path.join(directory, 'damage_bin_dict.bin')

    def generate_data(self):
        """
        Generate Damage Bin Dictionary dummy model file data.

        First bin always runs from 0 to 0, i.e. has a midpoint (interpolation)
        of 0. Last bin always runs from 0 to 0, i.e. has a midpoint
        (interpolation) of 1.

        Yields:
            bin_id (int): damage bin ID.
            bin_from (float): damage bin lower limit.
            bin_to (float): damage bin upper limit.
            interpolation (float): damage bin midpoint.
            interval_type (int): interval_type (deprecated).
        """
        # Exclude first and last bins for now
        bin_indexes = np.arange(self.num_damage_bins - 2)
        bin_from_values = bin_indexes / (self.num_damage_bins - 2)
        bin_to_values = (bin_indexes + 1) / (self.num_damage_bins - 2)
        # Set interpolation in middle of bin
        interpolations = (0.5 + bin_indexes) / (self.num_damage_bins - 2)
        # Insert first and last bins
        bin_indexes += 2
        bin_indexes = np.insert(bin_indexes, 0, 1)
        bin_indexes = np.append(bin_indexes, self.num_damage_bins)
        fields = [bin_from_values, bin_to_values, interpolations]
        for i, field in enumerate(fields):
            fields[i] = np.insert(field, 0, 0)
            fields[i] = np.append(fields[i], 1)
        bin_from_values, bin_to_values, interpolations = fields
        # Set interval type for all bins to 0 (unused)
        interval_type = 0

        for bin_id, bin_from, bin_to, interpolation in zip(
            bin_indexes, bin_from_values, bin_to_values, interpolations
        ):
            yield bin_id, bin_from, bin_to, interpolation, interval_type


class OccurrenceFile(ModelFile):
    """
    Generate data for Occurrence dummy model file.

    This file maps events to periods, which can represent any length of time.

    Attributes:
        get_num_periods_from_truncated_normal_cdf: Get number of periods on
            event-by-event basis.
        get_num_periods_from_truncated_normal_cdf: Get number of periods from
            truncated normal cumulative distribution function.
        set_occ_date_id: Set date of occurrence in ktools format.
        generate_data: Generate Occurrence dummy model file data.
    """

    def __init__(
            self, num_events, num_periods, random_seed, directory, mean, stddev
    ):
        """
        Initialise Occurrence file class.

        Args:
            num_events (int): number of events.
            num_periods (int): total number of periods.
            random_seed (float): random seed for random number generator.
            directory (str): dummy model file destination.
            mean (float): mean of truncated normal distribution sampled to
                determine number of periods per event.
            stddev (float): standard deviation of truncated normal distribution
                sampled to determine number of periods per event.
        """
        self.num_events = num_events
        self.num_periods = num_periods
        self.dtypes = OrderedDict([
            ('event_id', 'i'), ('period_no', 'i'), ('occ_date_id', 'i')
        ])
        self.date_algorithm = 1
        self.start_stats = [
            {
                'desc': 'Date algorithm', 'value': self.date_algorithm,
                'dtype': 'i'
            },
            {
                'desc': 'Number of periods', 'value': self.num_periods,
                'dtype': 'i'
            }
        ]
        self.random_seed = random_seed
        self.data_length = num_events
        self.mean = mean
        self.stddev = stddev
        self.file_name = os.path.join(directory, 'occurrence.bin')

    def get_num_periods_from_truncated_normal_cdf(self):
        """
        Get number of periods from truncated normal cumulative distribution
        function.

        Events can occur mupltiple times over multiple periods in the occurrence
        file. The number of periods per event is modelled by sampling from a
        truncated normal distribution with mean self.mean and standard deviation
        self.stddev. The lower tail of the distribution is truncated at 0.5 and
        the cumulative distribution function is given by:

        F(x) = [Phi(g(x)) - Phi(g(a))] / [Phi(g(b)) - Phi(g(a))]
        g(y) = (y - mean) / standard_deviation
        Phi(g(y)) = 1/2 * (1 + erf(g(y) / sqrt(2)))
        a = lower boundary = 0.5, b = upper boundary = infinity
          therefore g(b) -> infinity ===> Phi(g(b)) -> 1

        Returns:
            bound_a (int): lower boundary, when converted to an integer gives
                number of periods for this event.
        """
        alpha = (0.5 - self.mean) / self.stddev
        phi_alpha = 0.5 * (1 + erf(alpha / np.sqrt(2)))
        rand_no = np.random.random()
        bound_a = 0.5
        while True:
            xi = (bound_a - self.mean) / self.stddev
            phi_xi = 0.5 * (1 + erf(xi / np.sqrt(2)))
            if rand_no < ((phi_xi - phi_alpha) / (1 - phi_alpha)):
                return int(bound_a)
            bound_a += 1

    def get_num_periods_per_event(self):
        """
        Get number of periods on event-by-event basis.

        Determines whether sampling of truncated normal cumulative distribution
        function is required to obtain number of periods for this event.

        Returns:
            mean|bound_a (int): Number of periods for this event.
        """
        # Return mean if standard deviation is 0
        if self.stddev == 0:
            return self.mean
        else:
            return self.get_num_periods_from_truncated_normal_cdf()

    def set_occ_date_id(self, year, month, day):
        """
        Set date of occurrence in ktools format.

        Reduce year, month and day information to a single integer.

        Args:
            year (int): year.
            month (int): month.
            day (int): day.

        Returns:
            date (int): date in ktools format.
        """
        # Set date relative to epoch
        month = (month + 9) % 12
        year = year - month // 10
        return 365 * year + year // 4 - year // 100 + year // 400 + (306 * month + 5) // 10 + (day - 1)

    def generate_data(self):
        """
        Generate Occurrence dummy model file data.

        Yields:
            event (int): event ID.
            period_no (int): period number.
            date (int): date in ktools format.
        """
        super().seed_rng()
        months = np.arange(1, 13)
        days_per_month = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        months_weights = np.array(days_per_month, dtype=float)
        months_weights /= months_weights.sum()   # Normalise
        for event in range(self.num_events):
            for _ in range(self.get_num_periods_per_event()):
                period_no = np.random.randint(1, self.num_periods + 1)
                occ_year = period_no   # Assume one period represents one year
                occ_month = np.random.choice(months, p=months_weights)
                occ_day = np.random.randint(1, days_per_month[occ_month - 1])
                occ_date = self.set_occ_date_id(occ_year, occ_month, occ_day)
                yield event + 1, period_no, occ_date


class RandomFile(ModelFile):
    """
    Generate data for Random Numbers dummy model file.

    This optional file contains random numbers for ground up loss sampling.

    Attributes:
        generate_data: Generate Random Numbers dummy model file data.
    """

    def __init__(self, num_randoms, random_seed, directory):
        """
        Initialise Random Numbers file class.

        Args:
            num_randoms (int): number of random numbers.
            random_seed (float): random seed for random number generator.
            directory (str): dummy model file destination.
        """
        self.num_randoms = num_randoms
        self.dtypes = {'random_no': 'f'}
        self.start_stats = None
        self.random_seed = random_seed
        self.data_length = num_randoms
        self.file_name = os.path.join(directory, 'random.bin')

    def generate_data(self):
        """
        Generate Random Numbers dummy model file data.

        Yields:
            random number (float): random number.
        """
        super().seed_rng()
        # First random number is 0
        return (tuple([np.random.uniform()]) if i != 0 else (0,) for i in range(self.num_randoms))


class CoveragesFile(ModelFile):
    """
    Generate data for Coverages dummy model Oasis file.

    This file maps coverage IDs to Total Insured Values.

    Attributes:
        generate_data: Generate Coverages dummy model Oasis file data.
    """

    def __init__(
        self, num_locations, coverages_per_location, random_seed, directory
    ):
        """
        Initialise Coverages file class.

        Args:
            num_locations (int): number of locations.
            coverages_per_location (int): number of coverage types per location.
            random_seed (float): random seed for random number generator.
            directory (str): dummy model file destination.
        """
        self.num_locations = num_locations
        self.coverages_per_location = coverages_per_location
        self.dtypes = {'tiv': 'f'}
        self.start_stats = None
        self.random_seed = random_seed
        self.data_length = num_locations * coverages_per_location
        self.file_name = os.path.join(directory, 'coverages.bin')

    def generate_data(self):
        """
        Generate Coverages dummy model file data.

        Yields:
            total insured value (float): Total Insured Value (TIV).
        """
        super().seed_rng()
        # Assume 1-1 mapping between item and coverage IDs
        return (
            tuple([np.random.uniform(1, 1000000)]) for _ in range(
                self.num_locations * self.coverages_per_location
            )
        )


class ItemsFile(ModelFile):
    """
    Generate data for Items dummy model Oasis file.

    This file lists the exposure items for which ground up loss will be sampled.

    Attributes:
        generate_data: Generate Items dummy model Oasis file data.
    """

    def __init__(
        self, num_locations, coverages_per_location, num_areaperils,
        num_vulnerabilities, random_seed, directory
    ):
        """
        Initialise Items file class.

        Args:
            num_locations (int): number of locations.
            coverages_per_location (int): number of coverage types per location.
            num_areaperils (int): number of areaperils.
            num_vulnerabilities (int): number of vulnerabilities.
            random_seed (float): random seed for random number generator.
            directory (str): dummy model file destination.
        """
        self.num_locations = num_locations
        self.coverages_per_location = coverages_per_location
        self.num_areaperils = num_areaperils
        self.num_vulnerabilities = num_vulnerabilities
        self.dtypes = OrderedDict([
            ('item_id', 'i'), ('coverage_id', 'i'), ('areaperil_id', 'i'),
            ('vulnerability_id', 'i'), ('group_id', 'i'),
        ])
        self.start_stats = None
        self.random_seed = random_seed
        self.data_length = num_locations * coverages_per_location
        self.file_name = os.path.join(directory, 'items.bin')

    def generate_data(self):
        """
        Generate Items dummy model file data.

        Yields:
            item (int): item ID.
            item (int): coverage ID = item ID (1-1 mapping).
            areaperils[coverage] (int): areaperil ID corresponding to
                coverage ID.
            vulnerabilities[coverage] (int): vulnerability ID corresponding to
                coverage ID.
            location (int): group ID mapped to location ID.
        """
        super().seed_rng()
        for location in range(self.num_locations):
            areaperils = np.random.randint(
                1, self.num_areaperils + 1, size=self.coverages_per_location
            )
            vulnerabilities = np.random.randint(
                1, self.num_vulnerabilities + 1, size=self.coverages_per_location
            )
            for coverage in range(self.coverages_per_location):
                item = self.coverages_per_location * location + coverage + 1
                # Assume 1-1 mapping between item and coverage IDs
                # Assume group ID mapped to location
                yield item, item, areaperils[coverage], vulnerabilities[coverage], location + 1


class FMFile(ModelFile):
    """
    Parent class for generating random data for Financial Model files.
    """

    def __init__(self, num_locations, coverages_per_location):
        """
        Initialise Financial Model files classes.

        Args:
            num_locations (int): number of locations.
            coverages_per_location (int): number of coverage types per location.
        """
        self.num_locations = num_locations
        self.coverages_per_location = coverages_per_location
        self.start_stats = None


class FMProgrammeFile(FMFile):
    """
    Generate data for Financial Model Programme dummy model Oasis file.

    This file shows the level hierarchy.

    Attributes:
        generate_data: Generate Financial Model Programme dummy model Oasis file
            data.
    """

    def __init__(self, num_locations, coverages_per_location, directory):
        """
        Initialise Financial Model Programme file class.

        Args:
            num_locations (int): number of locations.
            coverages_per_location (int): number of coverage types per location.
            directory (str): dummy model file destination.
        """
        super().__init__(num_locations, coverages_per_location)
        self.dtypes = OrderedDict([
            ('from_agg_id', 'i'), ('level_id', 'i'), ('to_agg_id', 'i')
        ])
        self.data_length = num_locations * coverages_per_location * 2   # 2 from number of levels
        self.file_name = os.path.join(directory, 'fm_programme.bin')

    def generate_data(self):
        """
        Generate Financial Model Programme dummy model file data.

        Yields:
            agg_id (int): from aggregate ID.
            level (int): level ID.
            agg_id (int): to aggregate ID.
        """
        levels = [1, 10]
        levels = range(1, len(levels) + 1)
        for level in levels:
            for agg_id in range(
                1, self.num_locations * self.coverages_per_location + 1
            ):
                # Site coverage FM level
                if level == 1:
                    yield agg_id, level, agg_id
                # Policy layer FM level
                elif level == len(levels):
                    yield agg_id, level, 1


class FMPolicyTCFile(FMFile):
    """
    Generate data for Financial Model Policy dummy model Oasis file.

    This file shows the calculation rule (from the Financial Model Policy file)
    that should be applied to aggregations of loss at a particular level.

    Attributes:
        generate_data: Generate Financial Model Policy dummy model Oasis file
            data.
    """

    def __init__(
        self, num_locations, coverages_per_location, num_layers, directory
    ):
        """
        Initialise Financial Model Policy file class.

        Args:
            num_locations (int): number of locations.
            coverages_per_location (int): number of coverage types per location.
            num_layers (int): number of layers.
            directory (str): dummy model file destination.
        """
        super().__init__(num_locations, coverages_per_location)
        self.num_layers = num_layers
        self.dtypes = OrderedDict([
            ('layer_id', 'i'), ('level_id', 'i'), ('agg_id', 'i'),
            ('policytc_id', 'i')
        ])
        self.data_length = num_locations * coverages_per_location + num_layers
        self.file_name = os.path.join(directory, 'fm_policytc.bin')

    def generate_data(self):
        """
        Generate Financial Model Policy dummy model file data.

        Yields:
            level (int): level ID.
            agg_id (int): aggregate ID.
            layer (int): layer ID.
            policytc_id (int): profile/policyTC ID.
        """
        # Site coverage #1 & policy layer #10 FM levels
        levels = [1, 10]
        levels = range(1, len(levels) + 1)
        policytc_id = 1
        for level in levels:
            # Site coverage FM level
            if level == 1:
                for agg_id in range(
                    1, self.num_locations * self.coverages_per_location + 1
                ):
                    # One layer in site coverage FM level
                    yield level, agg_id, 1, policytc_id
                policytc_id += 1   # Next policytc_id
            # Policy layer FM level
            elif level == len(levels):
                for layer in range(self.num_layers):
                    yield level, 1, layer + 1, policytc_id
                    policytc_id += 1   # Next policytc_id


class FMProfileFile(ModelFile):
    """
    Generate data for Financial Model Profile dummy model Oasis file.

    This file contains the list of calculation rules with profile values used
    to generate insurance losses.

    Attributes:
        generate_data: Generate Financial Model Profile dummy model Oasis file
            data.
    """

    def __init__(self, num_layers, directory):
        """
        Initialise Financial Model Profile file class.

        Args:
            num_layers (int): number of layers.
            directory (str): dummy model file destination.
        """
        self.num_layers = num_layers
        self.dtypes = OrderedDict([
            ('policytc_id', 'i'), ('calcrule_id', 'i'), ('deductible1', 'f'),
            ('deductible2', 'f'), ('deductible3', 'f'), ('attachment1', 'f'),
            ('limit1', 'f'), ('share1', 'f'), ('share2', 'f'), ('share3', 'f')
        ])
        self.start_stats = None
        self.data_length = 1 + num_layers   # 1 from pass through at level 1
        self.file_name = os.path.join(directory, 'fm_profile.bin')

    def generate_data(self):
        """
        Generate Financial Model Profile dummy model file data.

        Yields:
            policytc_id (int): profile/policyTC ID.
            calculation rule ID (int): calculation rule ID (2 or 100).
            first deductible (float): first deductible (fixed at 0.0).
            second deductible (float): second deductible (fixed at 0.0).
            third deductible (float): third deductible (fixed at 0.0).
            attachment1 (float): attachment point/excess.
            limit1 (float): limit.
            first proportional share (float): first proportional
                share (0.0 or 0.3).
            second proportional share (float): second proportional
                share (fixed at 0.0).
            third proportional share (float): third proportional
                share (fixed at 0.0).
        """
        # Pass through for level 1
        profile_rows = [(1, 100, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)]
        # First policy
        init_policytc_id = 2
        init_attachment1 = 500000.0
        attachment1_offset = 5000000.0
        max_limit1 = 100000000.0
        for layer in range(self.num_layers):
            policytc_id = init_policytc_id + layer
            attachment1 = init_attachment1 + attachment1_offset * layer
            # Set limit1 at maximum for last layer
            if (layer + 1) == self.num_layers:
                limit1 = max_limit1
            else:
                limit1 = attachment1_offset * (layer + 1)
            profile_rows.append(
                (policytc_id, 2, 0.0, 0.0, 0.0, attachment1, limit1, 0.3, 0.0, 0.0)
            )
        for row in profile_rows:
            yield row


class FMXrefFile(FMFile):
    """
    Generate data for Financial Model Cross Reference dummy model Oasis file.

    This file shows the mapping between the financial model output ID, and
    aggregate and layer IDs.

    Attributes:
        generate_data: Generate Financial Model Cross Reference dummy model
            Oasis file data.
    """

    def __init__(
        self, num_locations, coverages_per_location, num_layers, directory
    ):
        """
        Initialise Financial Model Cross Reference file class.

        Args:
            num_locations (int): number of locations.
            coverages_per_location (int): number of coverage types per location.
            num_layers (int): number of layers.
            directory (str): dummy model file destination.
        """
        super().__init__(num_locations, coverages_per_location)
        self.num_layers = num_layers
        self.dtypes = OrderedDict([
            ('output', 'i'), ('agg_id', 'i'), ('layer_id', 'i')
        ])
        self.data_length = num_locations * coverages_per_location * num_layers
        self.file_name = os.path.join(directory, 'fm_xref.bin')

    def generate_data(self):
        """
        Generate Financial Model Cross Reference dummy model file data.

        Yields:
            output_count (int): output ID.
            agg_id (int): aggregate ID.
            layer (int): layer ID.
        """
        layers = range(1, self.num_layers + 1)
        output_count = 1
        for agg_id in range(
            1, self.num_locations * self.coverages_per_location + 1
        ):
            for layer in layers:
                yield output_count, agg_id, layer
                output_count += 1


class GULSummaryXrefFile(FMFile):
    """
    Generate data for Ground Up Losses Summary Cross Reference dummy model Oasis
    file.

    This file shows how item ground up losses are summed together at various
    summary levels in summarycalc.

    Attributes:
        generate_data: Generate Ground Up Losses Summary Cross Reference dummy
            model Oasis file data.
    """

    def __init__(self, num_locations, coverages_per_location, directory):
        """
        Initialise Ground Up Losses Summary Cross Reference file class.

        Args:
            num_locations (int): number of locations.
            coverages_per_location (int): number of coverage types per location.
            directory (str): dummy model file destination.
        """
        super().__init__(num_locations, coverages_per_location)
        self.dtypes = OrderedDict([
            ('item_id', 'i'), ('summary_id', 'i'), ('summaryset_id', 'i')
        ])
        self.data_length = num_locations * coverages_per_location
        self.file_name = os.path.join(directory, 'gulsummaryxref.bin')

    def generate_data(self):
        """
        Generate Ground Up Losses Summary Cross Reference dummy model file data.

        Yields:
            item (int): item ID.
            summary_id (int): summary ID.
            summaryset_id (int): summary set ID.
        """
        summary_id = 1
        summaryset_id = 1
        for item in range(self.num_locations * self.coverages_per_location):
            yield item + 1, summary_id, summaryset_id


class FMSummaryXrefFile(FMFile):
    """
    Generate data for Financial Model Summary Cross Reference dummy model Oasis
    file.

    This file shows how insurance losses are summed together at various levels
    by summarycalc.

    Attributes:
        generate_data: Generate Financial Model Summary Cross Reference dummy
            model Oasis file data.
    """

    def __init__(
        self, num_locations, coverages_per_location, num_layers, directory
    ):
        """
        Initialise Financial Model Summary Cross Reference file class.

        Args:
            num_locations (int): number of locations.
            coverages_per_location (int): number of coverage types per location.
            num_layers (int): number of layers.
            directory (str): dummy model file destination.
        """
        super().__init__(num_locations, coverages_per_location)
        self.num_layers = num_layers
        self.dtypes = OrderedDict([
            ('output_id', 'i'), ('summary_id', 'i'), ('summaryset_id', 'i')
        ])
        self.data_length = num_locations * coverages_per_location * num_layers
        self.file_name = os.path.join(directory, 'fmsummaryxref.bin')

    def generate_data(self):
        """
        Generate Financial Model Summary Cross Reference dummy model file data.

        Yields:
            output_id (int): output ID.
            summary_id (int): summary ID.
            summaryset_id (int): summary set ID.
        """
        summary_id = 1
        summaryset_id = 1
        for output_id in range(
            self.num_locations * self.coverages_per_location * self.num_layers
        ):
            yield output_id + 1, summary_id, summaryset_id



import logging
from numba import njit
from numba.core import types
from numba.typed import Dict
import numpy as np
import os

from .common import (
    BUFFER_SIZE,
    DATA_SIZE,
    N_PAIRS,
    FILE_HEADER_SIZE,
    event_count_dtype,
    amp_factor_dtype,
    AMPLIFICATIONS_FILE_NAME,
    LOSS_FACTORS_FILE_NAME
)

logger = logging.getLogger(__name__)


def get_items_amplifications(path):
    """
    Get array of amplification IDs from amplifications.bin, where index
    corresponds to item ID.

    amplifications.bin is binary file with layout:
        reserved header (4-byte int),
        item ID 1 (4-byte int), amplification ID a_1 (4-byte int),
        ...
        item ID n (4-byte int), amplification ID a_n (4-byte int)

    Args:
        path (str): path to amplifications.bin file

    Returns:
        items_amps (numpy.ndarray): array of amplification IDs, where index
            corresponds to item ID
    """
    items_amps = np.fromfile(
        os.path.join(path, AMPLIFICATIONS_FILE_NAME), dtype=np.int32,
        offset=FILE_HEADER_SIZE
    )

    # Check item IDs start from 1 and are contiguous
    if items_amps[0] != 1:
        logger.error(f'First item ID is {items_amps[0]}. Expected 1.')
        raise SystemExit(1)
    items_amps = items_amps.reshape(len(items_amps) // 2, 2)
    if not np.all(items_amps[1:, 0] - items_amps[:-1, 0] == 1):
        logger.error(f'Item IDs in {os.path.join(path, AMPLIFICATIONS_FILE_NAME)} are not contiguous')
        raise SystemExit(1)

    items_amps = np.concatenate((np.array([0]), items_amps[:, 1]))

    return items_amps


@njit(cache=True)
def fill_post_loss_amplification_factors(
    event_id, count, cursor, valid_length, event_count, amp_factor, plafactors
):
    """
    Fill Post Loss Amplification (PLA) factors dictionary mapped to
    event ID-item ID pair.

    Args:
        event_id (int): current event ID
        count (int): number of remaning amplification IDs associated with
            current event ID
        cursor (int): position in buffer
        valid_length (int): length of buffer in 8-bit chunks
        event_count (numpy.ndarray): array of event ID-count pairs
        amp_factor (numpy.ndarray): array of amplification ID-loss pairs
        plafactors (numba.typed.typeddict.Dict): PLA factors dictionary

    Returns:
        event_id (int): current event ID
        count (int): number of remaining amplification IDs associated with
            current event ID
        plafactors (numba.typed.typeddict.Dict): PLA factors dictionary
    """
    while cursor < valid_length:

        if count == 0:
            event_id = event_count[cursor]['event_id']
            count = event_count[cursor]['count']
            cursor += 1

        else:
            amplification_id = amp_factor[cursor]['amplification_id']
            loss_factor = amp_factor[cursor]['factor']
            plafactors[(event_id, amplification_id)] = loss_factor
            cursor += 1
            count -= 1

    return event_id, count, plafactors


def get_post_loss_amplification_factors(path):
    """
    Get Post Loss Amplification (PLA) factors mapped to event ID-item ID pair.

    lossfactors.bin is binary file with layout:
        reserved header (4-byte int),
        event ID 1 (4-byte int), number of amplification IDs for event ID 1 (4-byte int),
        amplification ID 1 (4-byte int), loss factor for amplification ID 1 (4-byte float),
        ...
        amplification ID n (4-byte int), loss factor for amplification ID n (4-byte float),
        event ID 2 (4-byte int), number of amplification IDs for event ID 2 (4-byte int),
        ...
        event ID N (4-byte int), number of amplification IDs for event ID N (4-byte int),
        amplification ID 1 (4-byte int), loss factor for amplification ID 1 (4-byte float),
        ...
        amplification ID n (4-byte int), loss factor for amplification ID n (4-byte float)

    Args:
        path (str): path to lossfactors.bin file

    Returns:
        plafactors (dict): event ID-item ID pairs mapped to amplification IDs
    """
    plafactors = Dict.empty(
        key_type=types.UniTuple(types.int64, 2), value_type=types.float64
    )

    with open(os.path.join(path, LOSS_FACTORS_FILE_NAME), 'rb') as f:
        factors_buffer = memoryview(bytearray(BUFFER_SIZE))
        event_count = np.ndarray(
            N_PAIRS, buffer=factors_buffer, dtype=event_count_dtype
        )
        amp_factor = np.ndarray(
            N_PAIRS, buffer=factors_buffer, dtype=amp_factor_dtype
        )
        f.readinto(factors_buffer[:FILE_HEADER_SIZE])   # Ignore first 4 bytes

        cursor = 0
        valid_buffer = 0
        count = 0
        event_id = 0
        while True:
            len_read = f.readinto(factors_buffer[valid_buffer:])
            valid_buffer += len_read

            if len_read == 0:
                break

            valid_length = valid_buffer // DATA_SIZE

            event_id, count, plafactors = fill_post_loss_amplification_factors(
                event_id, count, cursor, valid_length, event_count, amp_factor,
                plafactors
            )

            valid_buffer = 0
            cursor = 0

    return plafactors



from contextlib import ExitStack
import os
import sys

from .streams import read_and_write_streams
from .structure import (
    get_items_amplifications,
    get_post_loss_amplification_factors
)
from oasislmf.pytools.utils import redirect_logging


@redirect_logging(exec_name='plapy')
def run(run_dir, file_in, file_out, input_path, static_path):
    """
    Execeute the main Post Loss Amplification workflow.

    Args:
        run_dir (str): the directory of where the process is running
        file_in (str): file name of input stream
        file_out (str): file name of output streak
        input_path (str): path to amplifications.bin
        static_path (Str): path to lossfactors.bin

    Returns:
        0 (int): if no errors occurred
    """
    input_path = os.path.join(run_dir, input_path)
    static_path = os.path.join(run_dir, static_path)

    items_amps = get_items_amplifications(input_path)
    plafactors = get_post_loss_amplification_factors(static_path)

    with ExitStack() as stack:
        if file_in is None:
            stream_in = sys.stdin.buffer
        else:
            stream_in = stack.enter_context(open(file_in, 'rb'))

        if file_out is None:
            stream_out = sys.stdout.buffer
        else:
            stream_out = stack.enter_context(open(file_out, 'wb'))

        read_and_write_streams(stream_in, stream_out, items_amps, plafactors)

    return 0



"""
TODO: should I check values are valid in the financial structure percentage not between 0 and 1 (ex: deductible, limit ...)
TODO: validate max and min ded implementation
TODO: It seems that if a policy with share is used, subsequent policy using min or max deductible will be wrong
     so it make no sense to compute deductible, over_limit, under_limit

"""


from numba import njit
from .policy import (calcrule_28 as _calcrule_28, calcrule_32 as _calcrule_32, calcrule_34 as _calcrule_34,
                     calcrule_37 as _calcrule_37, calcrule_38 as _calcrule_38)


class UnknownCalcrule(Exception):
    pass


@njit(cache=True)
def min2(a, b):
    return a if a < b else b


@njit(cache=True)
def deductible_over_max(i, loss_out, loss_in, deductible, over_limit, under_limit, max_deductible):
    """
    deductible is over maximum deductible, we reduce the loss, therefore increase the loss up to under_limit

    under limit is always the minimum between the limit - loss and  the sub_node under_limit + the applied deductible
    so we are sure that if deductible[i] > max_ded_left, we are sure that under_limit is the good cap
    we are sure that is there is no sub node with limit loss_delta < under_limit

    loss delta can be negative so in this case we have to be careful it is not bigger than loss_in

    """
    loss_delta = deductible[i] - max_deductible
    if loss_delta > under_limit[i]:
        loss_out[i] = loss_in[i] + under_limit[i]
        over_limit[i] += loss_delta - under_limit[i]
        deductible[i] = max_deductible
        under_limit[i] = 0
    elif loss_in[i] >= -loss_delta:
        loss_out[i] = loss_in[i] + loss_delta
        under_limit[i] -= loss_delta
        deductible[i] = max_deductible
    else:
        loss_out[i] = 0
        deductible[i] += loss_in[i]
        under_limit[i] += loss_in[i]


@njit(cache=True)
def deductible_under_min(i, loss_out, loss_in, effective_deductible, over_limit, under_limit, min_deductible, deductible):
    """
    Deductible is under the minimum, we raise the deductible from over_limit first then loss if over_limit is not enough

    """
    loss_delta = min_deductible - deductible - effective_deductible[i]
    if loss_delta <= over_limit[i]:  # we have enough over_limit to cover loss_delta
        if loss_in[i] > deductible:  # we have enough loss to cover deductible
            loss_out[i] = loss_in[i]
            over_limit[i] -= loss_delta
            effective_deductible[i] = min_deductible
        elif (over_limit[i] - loss_delta) + loss_in[i] > deductible:  # not enough loss, we also reduce the over_limit
            loss_out[i] = 0
            over_limit[i] -= loss_delta + deductible
            effective_deductible[i] = min_deductible
            under_limit[i] += loss_in[i]
        else:
            effective_deductible[i] += loss_in[i] + over_limit[i]
            loss_out[i] = 0
            over_limit[i] = 0
            under_limit[i] += loss_in[i]

    else:
        loss_not_over_limit = loss_delta - over_limit[i]
        if loss_in[i] > loss_not_over_limit + deductible:  # we have enough loss after we use over_limit pool
            loss_out[i] = loss_in[i] - loss_not_over_limit - deductible
            over_limit[i] = 0
            effective_deductible[i] = min_deductible
            under_limit[i] += loss_not_over_limit
        else:
            loss_out[i] = 0
            effective_deductible[i] += loss_in[i] + over_limit[i]
            under_limit[i] += loss_in[i]
            over_limit[i] = 0


@njit(cache=True, fastmath=True)
def calcrule_1(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    Deductible and limit
    """
    lim = policy['limit_1'] + policy['deductible_1']
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= policy['deductible_1']:
            under_limit[i] = min2(under_limit[i] + loss_in[i], policy['limit_1'])
            deductible[i] += loss_in[i]
            loss_out[i] = 0
        elif loss_in[i] <= lim:
            under_limit[i] = min2(under_limit[i] + policy['deductible_1'], lim - loss_in[i])
            deductible[i] += policy['deductible_1']
            loss_out[i] = loss_in[i] - policy['deductible_1']
        else:
            over_limit[i] += loss_in[i] - lim
            under_limit[i] = 0
            deductible[i] += policy['deductible_1']
            loss_out[i] = policy['limit_1']


@njit(cache=True, fastmath=True)
def calcrule_2(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    Deductible, attachment, limit and share

    """
    ded_att = policy['deductible_1'] + policy['attachment_1']
    lim = policy['limit_1'] + ded_att
    maxi = policy['limit_1'] * policy['share_1']
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= ded_att:
            loss_out[i] = 0
        elif loss_in[i] <= lim:
            loss_out[i] = (loss_in[i] - ded_att) * policy['share_1']
        else:
            loss_out[i] = maxi


@njit(cache=True, fastmath=True)
def calcrule_3(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    Franchise deductible and limit
    """
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= policy['deductible_1']:
            under_limit[i] = min2(under_limit[i] + loss_in[i], policy['limit_1'])
            deductible[i] += loss_in[i]
            loss_out[i] = 0
        elif loss_in[i] <= policy['limit_1']:
            under_limit[i] = min2(under_limit[i], policy['limit_1'] - loss_in[i])
            loss_out[i] = loss_in[i]
        else:
            under_limit[i] = 0
            over_limit[i] += loss_in[i] - policy['limit_1']
            loss_out[i] = policy['limit_1']


@njit(cache=True, fastmath=True)
def calcrule_5(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    Deductible and limit as a proportion of loss
    """
    effective_deductible = loss_in * policy['deductible_1']
    effective_limit = loss_in * policy['limit_1']
    deductible += effective_deductible
    if policy['deductible_1'] + policy['limit_1'] >= 1:  # always under limit
        for i in range(loss_in.shape[0]):
            loss_out[i] = loss_in[i] - effective_deductible[i]
            under_limit[i] = min2(effective_limit[i] - loss_out[i], under_limit[i] + effective_deductible[i])

    else:  # always over limit
        loss_out[:] = effective_limit
        over_limit += loss_in - effective_deductible - effective_limit
        under_limit[:] = 0


@njit(cache=True, fastmath=True)
def calcrule_7(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible, minimum and maximum deductible, with limit
    """

    max_ded_left = policy['deductible_3'] - policy['deductible_1']
    min_ded_left = policy['deductible_2'] - policy['deductible_1']

    for i in range(loss_in.shape[0]):
        if deductible[i] > max_ded_left:
            deductible_over_max(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_3'])
        elif deductible[i] < min_ded_left:
            deductible_under_min(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_2'], policy['deductible_1'])
        else:
            if loss_in[i] > policy['deductible_1']:
                loss_out[i] = loss_in[i] - policy['deductible_1']
                deductible[i] += policy['deductible_1']
                under_limit[i] += policy['deductible_1']
            else:
                loss_out[i] = 0
                deductible[i] += loss_in[i]
                under_limit[i] += loss_in[i]

        if loss_out[i] > policy['limit_1']:
            over_limit[i] += loss_out[i] - policy['limit_1']
            under_limit[i] = 0
            loss_out[i] = policy['limit_1']
        else:
            under_limit[i] = min2(policy['limit_1'] - loss_out[i], under_limit[i])


@njit(cache=True, fastmath=True)
def calcrule_8(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible and minimum deductible, with limit
    """
    min_ded_left = policy['deductible_2'] - policy['deductible_1']
    for i in range(loss_in.shape[0]):
        if deductible[i] < min_ded_left:
            deductible_under_min(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_2'], policy['deductible_1'])
        else:
            if loss_in[i] > policy['deductible_1']:
                loss_out[i] = loss_in[i] - policy['deductible_1']
                deductible[i] += policy['deductible_1']
                under_limit[i] += policy['deductible_1']
            else:
                loss_out[i] = 0
                deductible[i] += loss_in[i]
                under_limit[i] += loss_in[i]

        if loss_out[i] > policy['limit_1']:
            over_limit[i] += loss_out[i] - policy['limit_1']
            under_limit[i] = 0
            loss_out[i] = policy['limit_1']
        else:
            under_limit[i] = min2(policy['limit_1'] - loss_out[i], under_limit[i])


@njit(cache=True, fastmath=True)
def calcrule_10(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible and maximum deductible
    """
    max_ded_left = policy['deductible_3'] - policy['deductible_1']

    for i in range(loss_in.shape[0]):
        if deductible[i] > max_ded_left:
            deductible_over_max(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_3'])
        else:
            if loss_in[i] > policy['deductible_1']:
                loss_out[i] = loss_in[i] - policy['deductible_1']
                deductible[i] += policy['deductible_1']
                under_limit[i] += policy['deductible_1']
            else:
                loss_out[i] = 0
                deductible[i] += loss_in[i]
                under_limit[i] += loss_in[i]


@njit(cache=True, fastmath=True)
def calcrule_11(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible and minimum deductible
    """
    min_ded_left = policy['deductible_2'] - policy['deductible_1']

    for i in range(loss_in.shape[0]):
        if deductible[i] < min_ded_left:
            deductible_under_min(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_2'], policy['deductible_1'])
        else:
            if loss_in[i] > policy['deductible_1']:
                loss_out[i] = loss_in[i] - policy['deductible_1']
                deductible[i] += policy['deductible_1']
                under_limit[i] += policy['deductible_1']
            else:
                loss_out[i] = 0
                deductible[i] += loss_in[i]
                under_limit[i] += loss_in[i]


@njit(cache=True, fastmath=True)
def calcrule_12(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    Deductible only
    """
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= policy['deductible_1']:
            under_limit[i] += loss_in[i]
            deductible[i] += loss_in[i]
            loss_out[i] = 0
        else:
            under_limit[i] += policy['deductible_1']
            deductible[i] += policy['deductible_1']
            loss_out[i] = loss_in[i] - policy['deductible_1']


@njit(cache=True, fastmath=True)
def calcrule_13(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible, minimum and maximum deductible
    """

    max_ded_left = policy['deductible_3'] - policy['deductible_1']
    min_ded_left = policy['deductible_2'] - policy['deductible_1']

    for i in range(loss_in.shape[0]):
        if deductible[i] > max_ded_left:
            deductible_over_max(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_3'])
        elif deductible[i] < min_ded_left:
            deductible_under_min(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_2'], policy['deductible_1'])
        else:
            if loss_in[i] > policy['deductible_1']:
                loss_out[i] = loss_in[i] - policy['deductible_1']
                deductible[i] += policy['deductible_1']
                under_limit[i] += policy['deductible_1']
            else:
                loss_out[i] = 0
                deductible[i] += loss_in[i]
                under_limit[i] += loss_in[i]


@njit(cache=True, fastmath=True)
def calcrule_14(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    Limit only
    """
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= policy['limit_1']:
            under_limit[i] = min2(policy['limit_1'] - loss_in[i], under_limit[i])
            loss_out[i] = loss_in[i]
        else:
            over_limit[i] += loss_in[i] - policy['limit_1']
            under_limit[i] = 0
            loss_out[i] = policy['limit_1']


@njit(cache=True, fastmath=True)
def calcrule_15(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible and limit % loss
    """
    effective_limit = policy['deductible_1'] / (1 - policy['limit_1'])
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= policy['deductible_1']:
            under_limit[i] = min2(effective_limit, under_limit[i] + loss_in[i])
            loss_out[i] = 0
            deductible[i] += loss_in[i]
        elif loss_in[i] <= effective_limit:
            under_limit[i] = min2(effective_limit - loss_in[i], under_limit[i] + policy['deductible_1'])
            loss_out[i] = loss_in[i] - policy['deductible_1']
            deductible[i] += policy['deductible_1']
        else:
            loss_out[i] = loss_in[i] * policy['limit_1']
            deductible[i] += policy['deductible_1']
            over_limit[i] += loss_in[i] - loss_out[i] - policy['deductible_1']
            under_limit[i] = 0


@njit(cache=True, fastmath=True)
def calcrule_16(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible % loss
    """
    effective_deductible = loss_in * policy['deductible_1']
    deductible += effective_deductible
    under_limit += effective_deductible
    loss_out[:] = loss_in - effective_deductible


@njit(cache=True, fastmath=True)
def calcrule_17(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible % loss with attachment, limit and share
    """
    if policy['deductible_1'] >= 1:
        loss_out.fill(0)
    else:
        post_ded_attachment = policy['attachment_1'] / (1 - policy['deductible_1'])
        post_ded_attachment_limit = (policy['attachment_1'] + policy['limit_1']) / (1 - policy['deductible_1'])
        maxi = policy['limit_1'] * policy['share_1']
        for i in range(loss_in.shape[0]):
            effective_deductible = loss_in[i] * policy['deductible_1']
            if loss_in[i] <= post_ded_attachment:
                loss_out[i] = 0
            elif loss_in[i] <= post_ded_attachment_limit:
                loss_out[i] = (loss_in[i] - effective_deductible - policy['attachment_1']) * policy['share_1']
            else:
                loss_out[i] = maxi


@njit(cache=True, fastmath=True)
def calcrule_19(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible % loss with min and/or max deductible

    TODO: check if we can assume 0 <= policy['deductible_1'] <= 1
    """

    for i in range(loss_in.shape[0]):
        effective_deductible = loss_in[i] * policy['deductible_1']
        if effective_deductible + deductible[i] > policy['deductible_3'] > 0:
            deductible_over_max(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_3'])
        elif effective_deductible + deductible[i] < policy['deductible_2']:
            deductible_under_min(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_2'], policy['deductible_1'])
        else:
            if loss_in[i] > effective_deductible:
                loss_out[i] = loss_in[i] - effective_deductible
                deductible[i] += effective_deductible
                under_limit[i] += effective_deductible
            else:
                loss_out[i] = 0
                deductible[i] += loss_in[i]
                under_limit[i] += loss_in[i]


@njit(cache=True, fastmath=True)
def calcrule_20(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    reverse franchise deductible
    """
    for i in range(loss_in.shape[0]):
        if loss_in[i] > policy['deductible_1']:
            loss_out[i] = 0
        else:
            loss_out[i] = loss_in[i]


@njit(cache=True, fastmath=True)
def calcrule_22(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    reinsurance % ceded, limit and % placed
    """
    if policy['share_1'] == 0:
        loss_out.fill(0)
    else:
        pre_share_limit = policy['limit_1'] / policy['share_1']
        all_share = policy['share_1'] * policy['share_2'] * policy['share_3']
        maxi = policy['limit_1'] * policy['share_2'] * policy['share_3']
        for i in range(loss_in.shape[0]):
            if loss_in[i] <= pre_share_limit:
                loss_out[i] = loss_in[i] * all_share
            else:
                loss_out[i] = maxi


@njit(cache=True, fastmath=True)
def calcrule_23(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    reinsurance limit and % placed
    """
    all_share = policy['share_2'] * policy['share_3']
    maxi = policy['limit_1'] * all_share
    for i in range(loss_in.shape[0]):
        if loss_in[i] <= policy['limit_1']:
            loss_out[i] = loss_in[i] * all_share
        else:
            loss_out[i] = maxi


@njit(cache=True, fastmath=True)
def calcrule_24(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    reinsurance excess terms
    """
    if policy['share_1'] == 0:
        loss_out.fill(0)
    else:
        pre_share_attachment = policy['attachment_1'] / policy['share_1']
        pre_share_attachment_limit = (policy['limit_1'] + policy['attachment_1']) / policy['share_1']
        attachment_share = policy['attachment_1'] * policy['share_2'] * policy['share_3']
        all_share = policy['share_1'] * policy['share_2'] * policy['share_3']
        maxi = policy['limit_1'] * policy['share_2'] * policy['share_3']
        for i in range(loss_in.shape[0]):
            if loss_in[i] <= pre_share_attachment:
                loss_out[i] = 0
            elif loss_in[i] <= pre_share_attachment_limit:
                loss_out[i] = loss_in[i] * all_share - attachment_share
            else:
                loss_out[i] = maxi


@njit(cache=True, fastmath=True)
def calcrule_25(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    reinsurance proportional terms
    """
    loss_out[:] = loss_in * (policy['share_1'] * policy['share_2'] * policy['share_3'])


@njit(cache=True, fastmath=True)
def calcrule_26(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible % loss with min and/or max deductible and limit

    TODO: check if we can assume 0 <= policy['deductible_1'] <= 1
    """

    for i in range(loss_in.shape[0]):
        effective_deductible = loss_in[i] * policy['deductible_1']
        if effective_deductible + deductible[i] > policy['deductible_3'] > 0:
            deductible_over_max(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_3'])
        elif effective_deductible + deductible[i] < policy['deductible_2']:
            deductible_under_min(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_2'], effective_deductible)
        else:
            if loss_in[i] > effective_deductible:
                loss_out[i] = loss_in[i] - effective_deductible
                deductible[i] += effective_deductible
                under_limit[i] += effective_deductible
            else:
                loss_out[i] = 0
                deductible[i] += loss_in[i]
                under_limit[i] += loss_in[i]

        if loss_out[i] > policy['limit_1']:
            over_limit[i] += loss_out[i] - policy['limit_1']
            under_limit[i] = 0
            loss_out[i] = policy['limit_1']
        else:
            under_limit[i] = min2(policy['limit_1'] - loss_out[i], under_limit[i])


@njit(cache=True, fastmath=True)
def calcrule_27(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    step payout with limit
    """
    if policy['step_id'] == 1:
        loss_out.fill(0)
    for i in range(loss_in.shape[0]):
        if (0 < loss_in[i] or 0 < deductible[i]) and policy['trigger_start'] <= loss_in[i] < policy['trigger_end']:
            loss_out[i] += policy['payout_start']


@njit(cache=True, fastmath=True)
def calcrule_28(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    % loss step payout
    """
    _calcrule_28(policy, loss_out, loss_in)


@njit(cache=True, fastmath=True)
def calcrule_32(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    monetary amount trigger and % loss step payout with limit
    """
    _calcrule_32(policy, loss_out, loss_in)


@njit(cache=True, fastmath=True)
def calcrule_33(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible % loss with limit

    """
    if policy['deductible_1'] >= 1:
        loss_out.fill(0)
        deductible += loss_in
    else:
        post_ded_limit = policy['limit_1'] / (1 - policy['deductible_1'])
        for i in range(loss_in.shape[0]):
            effective_deductible = loss_in[i] * policy['deductible_1']
            deductible[i] += effective_deductible
            if loss_in[i] <= post_ded_limit:
                loss_out[i] = loss_in[i] - effective_deductible
                under_limit[i] = min2(under_limit[i] + effective_deductible, policy['limit_1'] - loss_out[i])
            else:
                over_limit[i] += loss_in[i] - effective_deductible - policy['limit_1']
                under_limit[i] = 0
                loss_out[i] = policy['limit_1']


@njit(cache=True, fastmath=True)
def calcrule_34(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible with attachment and share

    TODO: compare to the cpp, as there is shares, deductible won't be use later on so no need to compute it
    """
    _calcrule_34(policy, loss_out, loss_in)


@njit(cache=True, fastmath=True)
def calcrule_35(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible % loss with min and/or max deductible and limit % loss

    TODO: check if we can assume 0 <= policy['deductible_1'] <= 1
    """

    for i in range(loss_in.shape[0]):
        effective_deductible = loss_in[i] * policy['deductible_1']
        if effective_deductible + deductible[i] > policy['deductible_3'] > 0:
            deductible_over_max(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_3'])
        elif effective_deductible + deductible[i] < policy['deductible_2']:
            deductible_under_min(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_2'], effective_deductible)
        else:
            if loss_in[i] > effective_deductible:
                loss_out[i] = loss_in[i] - effective_deductible
                deductible[i] += effective_deductible
                under_limit[i] += effective_deductible
            else:
                loss_out[i] = 0
                deductible[i] += loss_in[i]
                under_limit[i] += loss_in[i]
        limit = loss_in[i] * policy['limit_1']
        if loss_out[i] > limit:
            over_limit[i] += loss_out[i] - limit
            under_limit[i] = 0
            loss_out[i] = limit
        else:
            under_limit[i] = min2(limit - loss_out[i], under_limit[i])


@njit(cache=True, fastmath=True)
def calcrule_36(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    deductible with min and/or max deductible and limit % loss
    """

    max_ded_left = policy['deductible_3'] - policy['deductible_1']
    min_ded_left = policy['deductible_2'] - policy['deductible_1']

    for i in range(loss_in.shape[0]):
        if deductible[i] > max_ded_left > 0:
            deductible_over_max(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_3'])
        elif deductible[i] < min_ded_left:
            deductible_under_min(i, loss_out, loss_in, deductible, over_limit, under_limit, policy['deductible_2'], policy['deductible_1'])
        else:
            if loss_in[i] > policy['deductible_1']:
                loss_out[i] = loss_in[i] - policy['deductible_1']
                deductible[i] += policy['deductible_1']
                under_limit[i] += policy['deductible_1']
            else:
                loss_out[i] = 0
                deductible[i] += loss_in[i]
                under_limit[i] += loss_in[i]

        limit = loss_in[i] * policy['limit_1']
        if loss_out[i] > limit:
            over_limit[i] += loss_out[i] - limit
            under_limit[i] = 0
            loss_out[i] = limit
        else:
            under_limit[i] = min2(limit - loss_out[i], under_limit[i])


@njit(cache=True, fastmath=True)
def calcrule_37(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    % loss step payout
    """
    _calcrule_37(policy, loss_out, loss_in)
    if policy['step_id'] == 1:
        loss_out.fill(0)
    for i in range(loss_in.shape[0]):
        if policy['trigger_start'] <= loss_in[i] < policy['trigger_end']:
            loss = min(max(policy['payout_start'] * loss_in[i] - policy['deductible_1'], 0), policy['limit_1'])
            loss_out[i] = (loss + min(loss * policy['scale_2'], policy['limit_2'])) * policy['scale_1']


@njit(cache=True, fastmath=True)
def calcrule_38(policy, loss_out, loss_in, deductible, over_limit, under_limit):
    """
    conditional coverage
    """
    _calcrule_38(policy, loss_out, loss_in)


@njit(cache=True)
def calc(policy, loss_out, loss_in, deductible, over_limit, under_limit, stepped):
    if policy['calcrule_id'] == 1:
        calcrule_1(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 2:
        calcrule_2(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 3:
        calcrule_3(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    # calcrule_4 (deductible % TIV and limit) is redirected to 1 when building financial structure
    elif policy['calcrule_id'] == 5:
        calcrule_5(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    # calcrule_6 (deductible % TIV) is redirected to 12 when building financial structure
    elif policy['calcrule_id'] == 7:
        calcrule_7(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 8:
        calcrule_8(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    # calcrule_9 (limit with deductible % limit) is redirected to 1 when building financial structure
    elif policy['calcrule_id'] == 10:
        calcrule_10(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 11:
        calcrule_11(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 12:
        calcrule_12(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 13:
        calcrule_13(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 14:
        calcrule_14(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 15:
        calcrule_15(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 16:
        calcrule_16(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 17:
        calcrule_17(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    # calcrule_18 (deductible % tiv with attachment, limit and share) is redirected to 2 when building financial structure
    elif policy['calcrule_id'] == 19:
        calcrule_19(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 20:
        calcrule_20(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    # calcrule_21 (deductible % tiv with min and max deductible) is redirected to 13 when building financial structure
    elif policy['calcrule_id'] == 22:
        calcrule_22(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 23:
        calcrule_23(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 24:
        calcrule_24(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 25:
        calcrule_25(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 26:
        calcrule_26(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 33:
        calcrule_33(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 34:
        calcrule_34(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 35:
        calcrule_35(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 36:
        calcrule_36(policy, loss_out, loss_in, deductible, over_limit, under_limit)
    elif policy['calcrule_id'] == 100:
        loss_out[:] = loss_in
    elif stepped is not None:
        if policy['calcrule_id'] == 27:
            calcrule_27(policy, loss_out, loss_in, deductible, over_limit, under_limit)
        elif policy['calcrule_id'] == 28:
            calcrule_28(policy, loss_out, loss_in, deductible, over_limit, under_limit)
        elif policy['calcrule_id'] == 32:
            calcrule_32(policy, loss_out, loss_in, deductible, over_limit, under_limit)
        elif policy['calcrule_id'] == 37:
            calcrule_37(policy, loss_out, loss_in, deductible, over_limit, under_limit)
        elif policy['calcrule_id'] == 38:
            calcrule_38(policy, loss_out, loss_in, deductible, over_limit, under_limit)
        else:
            raise UnknownCalcrule()
    else:
        raise UnknownCalcrule()



import tempfile
import sys
import logging
import numpy as np

from .financial_structure import create_financial_structure, load_financial_structure
from .stream_sparse import read_stream_header, EventWriterSparse, read_streams_sparse, EventWriterOrderedOutputSparse
from .compute_sparse import compute_event as compute_event_sparse
from .compute_sparse import init_variable as init_variable_sparse
from .compute_sparse import reset_variable as reset_variable_sparse
from oasislmf.pytools.utils import redirect_logging


logger = logging.getLogger(__name__)


def run(create_financial_structure_files, **kwargs):
    if create_financial_structure_files:
        create_financial_structure(kwargs['allocation_rule'], kwargs['static_path'])
    else:
        return run_synchronous(**kwargs)


@redirect_logging(exec_name='fmpy')
def run_synchronous(allocation_rule, files_in, files_out, net_loss, storage_method, **kwargs):
    if allocation_rule == 3:
        allocation_rule = 2
    elif allocation_rule == 0 and net_loss:
        raise NotImplementedError("net loss option is not implemented for alloc rule 0")

    if files_out is not None:
        files_out = files_out[0]

    if files_in is None:
        streams_in = [sys.stdin.buffer]
    else:
        streams_in = [open(file_in, 'rb') for file_in in files_in]

    try:
        for stream_in in streams_in:
            stream_type, max_sidx_val = read_stream_header(stream_in)

        if storage_method == "sparse":
            run_synchronous_sparse(max_sidx_val, allocation_rule, streams_in=streams_in, files_out=files_out, net_loss=net_loss, **kwargs)
        else:
            raise ValueError(f"storage_method {storage_method} is not supported for this version")

    finally:
        if files_in is not None:
            for stream_in in streams_in:
                stream_in.close()


def run_synchronous_sparse(max_sidx_val, allocation_rule, static_path, streams_in, files_out, low_memory, net_loss, sort_output, **kwargs):
    compute_info, nodes_array, node_parents_array, node_profiles_array, output_array, fm_profile = load_financial_structure(
        allocation_rule, static_path)

    compute_info = compute_info[0]
    stepped = True if compute_info['stepped'] else None  # https://github.com/numba/numba/issues/4108

    if sort_output:
        event_writer_cls = EventWriterOrderedOutputSparse
    else:
        event_writer_cls = EventWriterSparse

    with tempfile.TemporaryDirectory() as tempdir:
        (max_sidx_val, max_sidx_count, len_array, sidx_indexes, sidx_indptr, sidx_val, loss_indptr, loss_val, pass_through,
         extras_indptr, extras_val, children, computes, item_parent_i, compute_idx) = init_variable_sparse(compute_info, max_sidx_val, tempdir, low_memory)

        if allocation_rule == 0:
            pass_through_out = np.zeros_like(pass_through)
        else:
            pass_through_out = pass_through

        with event_writer_cls(files_out, nodes_array, output_array, sidx_indexes, sidx_indptr, sidx_val,
                              loss_indptr, loss_val, pass_through_out, max_sidx_val, computes) as event_writer:
            for event_id in read_streams_sparse(streams_in, nodes_array, sidx_indexes, sidx_indptr, sidx_val,
                                                loss_indptr, loss_val, pass_through, len_array, computes, compute_idx):
                compute_event_sparse(
                    compute_info,
                    net_loss,
                    nodes_array,
                    node_parents_array,
                    node_profiles_array,
                    len_array, max_sidx_val, sidx_indexes, sidx_indptr, sidx_val, loss_indptr, loss_val, extras_indptr, extras_val,
                    children,
                    computes,
                    compute_idx,
                    item_parent_i,
                    fm_profile,
                    stepped)
                event_writer.write(event_id, compute_idx)
                reset_variable_sparse(children, compute_idx, computes)



"""
This file defines the functions that maps the supported perils with the correlation settings. This data is usually
obtained from the model_settings.
"""
from typing import Optional

import pandas as pd


def map_data(data: Optional[dict]) -> Optional[pd.DataFrame]:
    """
    Maps data from the model settings to to have Peril ID, peril_correlation_group, and damage_correlation_value.

    Args:
        data: (dict) the data loaded from the model settings

    Returns: (pd.DataFrame) the mapped data
    """
    if data is not None:
        supported_perils = data.get("lookup_settings", {}).get("supported_perils", [])
        correlation_settings = data.get("correlation_settings", [])

        for supported_peril in supported_perils:
            supported_peril["peril_correlation_group"] = supported_peril.get("peril_correlation_group", 0)

        supported_perils_df = pd.DataFrame(supported_perils)
        correlation_settings_df = pd.DataFrame(correlation_settings)

        if len(correlation_settings_df) > 0:
            # correlations_settings are defined
            if "damage_correlation_value" not in correlation_settings_df.columns:
                raise ValueError(
                    "Expect correlation settings in model settings file to contain a `damage_correlation_value` "
                    f"in each `peril_correlation_group` entry, got \n{correlation_settings}.")

            if "hazard_correlation_value" not in correlation_settings_df.columns:
                raise ValueError(
                    "Expect correlation settings in model settings file to contain a `hazard_correlation_value` "
                    f"in each `peril_correlation_group` entry, got \n {correlation_settings}.")

        # merge allows duplicates of the "peril_correlation_group" in the supported perils
        # merge does not allow duplicates of the "peril_correlation_group" in the correlation settings
        if len(supported_perils_df) > 0 and len(correlation_settings_df) > 0:
            mapped_data = pd.merge(supported_perils_df, correlation_settings_df, on="peril_correlation_group")
            return mapped_data


def get_correlation_input_items(gul_inputs_df: pd.DataFrame, correlation_map_df: pd.DataFrame) -> pd.DataFrame:
    """
    Gets the correlation values with the peril ID from the model_settings.

    Args:
        correlation_map_df: (pd.DataFrame) data from the model settings to to have Peril ID, peril_correlation_group,
                                           and damage_correlation_value
        gul_inputs_df: (pd.DataFrame) the data of the gul inputs to be mapped

    Returns: (pd.DataFrame) the mapped data of correlations
    """
    correlation_df = (
        gul_inputs_df
        .merge(correlation_map_df, left_on='peril_id', right_on='id')
        .reset_index()
        .astype({"damage_correlation_value": "float32", "hazard_correlation_value": "float32"})
        [["item_id", "peril_correlation_group", "damage_correlation_value", "hazard_group_id", "hazard_correlation_value"]]
        .sort_values('item_id')
    )

    return correlation_df



__all__ = [
    'as_path',
    'empty_dir',
    'PathCleaner',
    'import_from_string',
    'get_custom_module',
    'setcwd'
]

import os
import sys
import importlib
import re
import shutil

from contextlib import contextmanager

from .exceptions import OasisException


def as_path(path, label, is_dir=False, preexists=True, null_is_valid=True):
    """
    Processes the path and returns the absolute path.

    If the path does not exist and ``preexists`` is true
    an ``OasisException`` is raised.

    :param path: The path to process
    :type path: str

    :param label: Human-readable label of the path (used for error reporting)
    :type label: str

    :param is_dir: Whether the path is a directory
    :type is_dir: bool

    :param preexists: Flag whether to raise an error if the path
        does not exist.
    :type preexists: bool

    :param null_is_valid: flag to indicate if None is a valid value
    :type null_is_valid: bool

    :return: The absolute path of the input path
    """
    if path is None and null_is_valid:
        return

    if not isinstance(path, str):
        if preexists:
            raise OasisException(f'The path {path} ({label}) is indicated as preexisting but is not a valid path')
        else:
            return
    if not os.path.isabs(path):
        path = os.path.abspath(path)
    if preexists and not os.path.exists(path):
        raise OasisException(f'The path {path} ({label}) is indicated as preexisting but does not exist')
    if is_dir and preexists and not os.path.isdir(path):
        raise OasisException(f'The path {path} ({label}) is indicated as a preexisting directory but is not actually a directory')

    return os.path.normpath(path)


def empty_dir(dir_fp):
    """
    Empties the contents of a directory, but leaves the directory in place.

    :param dir_fp: A pre-existing directory path
    :type dir_fp: str
    """
    _dir_fp = as_path(dir_fp, dir_fp, is_dir=True)

    for p in (os.path.join(_dir_fp, fn) for fn in os.listdir(_dir_fp)):
        os.remove(p) if os.path.isfile(p) else (shutil.rmtree(p) if os.path.isdir(p) else None)


class PathCleaner(object):
    """
    A callable that generates the absolute path of the given path and checks
    that it exists if indicated as preexisting.

    :param label: A user-friendly label for the path (used for error reporting)
    :type label: str

    :param preexists: Flag whether to raise an error if the path
        does not exist.
    :type preexists: bool
    """

    def __init__(self, label, preexists=True):
        self.label = label
        self.preexists = preexists

    def __call__(self, path):
        return as_path(path, self.label, preexists=self.preexists)


def import_from_string(name):
    """
    return the object or module from the path given
    >>> import os.path
    >>> mod = import_from_string('os.path')
    >>> os.path is mod
    True

    >>> from os.path import isabs
    >>> cls = import_from_string('os.path.isabs')
    >>> isabs is cls
    True
    """
    components = name.split('.')
    res = __import__(components[0])
    for comp in components[1:]:
        res = getattr(res, comp)
    return res


def get_custom_module(custom_module_path, label):
    """
    return the custom module present at the custom_module_path.
    the try loop allow for the custom module to work even if it depends on other module of its package
    by testing recursively for the presence of __init__.py file

    (ex: this module "path" is using from .exceptions import OasisException so it can only be imported as part of the
    utils package => sys.path.insert(0, path_to_utils);importlib.import_module('utils.path'))
    >>> mod = get_custom_module(__file__, "test module")
    >>> mod.__name__.rsplit('.', 1)[-1]
    'path'
    """
    custom_module_path = as_path(custom_module_path, label, preexists=True, null_is_valid=False)

    package_dir = os.path.dirname(custom_module_path)
    module_name = re.sub(r'\.py$', '', os.path.basename(custom_module_path))

    while True:
        sys.path.insert(0, package_dir)
        try:
            custom_module = importlib.import_module(module_name)
            importlib.reload(custom_module)
            return custom_module
        except ImportError:
            if '__init__.py' in os.listdir(package_dir):
                module_name = os.path.basename(package_dir) + '.' + module_name
                package_dir, old_package_dir = os.path.dirname(package_dir), package_dir
                if package_dir == old_package_dir:
                    raise
            else:
                raise
        finally:
            sys.path.pop(0)


@contextmanager
def setcwd(path):
    orig_path = os.getcwd()
    try:
        os.chdir(str(path))
        yield path
    finally:
        os.chdir(orig_path)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



__all__ = [
    'get_gul_input_items',
    'write_amplifications_file',
    'write_coverages_file',
    'write_gul_input_files',
    'write_items_file',
    'write_complex_items_file'
]
import copy
import os
import sys
import warnings
from collections import OrderedDict

import pandas as pd

from oasislmf.pytools.data_layer.oasis_files.correlations import CorrelationsData
from oasislmf.utils.coverages import SUPPORTED_COVERAGE_TYPES
from oasislmf.utils.data import factorize_ndarray, merge_dataframes, set_dataframe_column_dtypes
from oasislmf.utils.defaults import (CORRELATION_GROUP_ID, DAMAGE_GROUP_ID_COLS,
                                     HAZARD_GROUP_ID_COLS, OASIS_FILES_PREFIXES, SOURCE_IDX,
                                     get_default_exposure_profile)
from oasislmf.utils.exceptions import OasisException
from oasislmf.utils.fm import SUPPORTED_FM_LEVELS
from oasislmf.utils.log import oasis_log
from oasislmf.utils.path import as_path
from oasislmf.utils.profiles import (get_fm_terms_oed_columns,
                                     get_grouped_fm_profile_by_level_and_term_group,
                                     get_grouped_fm_terms_by_level_and_term_group,
                                     get_oed_hierarchy)

pd.options.mode.chained_assignment = None
warnings.simplefilter(action='ignore', category=FutureWarning)

VALID_OASIS_GROUP_COLS = [
    'item_id',
    'peril_id',
    'coverage_id',
    'coverage_type_id',
    'peril_correlation_group'
]

PERIL_CORRELATION_GROUP_COL = 'peril_correlation_group'


def process_group_id_cols(group_id_cols, exposure_df_columns, has_correlation_groups):
    """
    cleans out columns that are not valid oasis group columns.

    Valid group id columns can be either
    1. exist in the location file
    2. be listed as a useful internal col

    Args:
        group_id_cols: (List[str]) the ID columns that are going to be filtered
        exposure_df_columns: (List[str]) the columns in the exposure dataframe
        has_correlation_groups: (bool) if set to True means that we are hashing with correlations in mind therefore the
                             "peril_correlation_group" column is added

    Returns: (List[str]) the filtered columns
    """
    for col in group_id_cols:
        if col not in list(exposure_df_columns) + VALID_OASIS_GROUP_COLS:
            warnings.warn('Column {} not found in loc file, or a valid internal oasis column'.format(col))
            group_id_cols.remove(col)

    if PERIL_CORRELATION_GROUP_COL not in group_id_cols and has_correlation_groups is True:
        group_id_cols.append(PERIL_CORRELATION_GROUP_COL)

    return group_id_cols


@oasis_log
def get_gul_input_items(
    location_df,
    keys_df,
    correlations=False,
    peril_correlation_group_df=None,
    exposure_profile=get_default_exposure_profile(),
    damage_group_id_cols=None,
    hazard_group_id_cols=None,
):
    """
    Generates and returns a Pandas dataframe of GUL input items.

    :param exposure_df: Exposure dataframe
    :type exposure_df: pandas.DataFrame

    :param keys_df: Keys dataframe
    :type keys_df: pandas.DataFrame

    :param output_dir: the output directory where input files are stored
    :type output_dir: str

    :param exposure_profile: Exposure profile
    :type exposure_profile: dict

    :param damage_group_id_cols: Columns to be used to generate a hashed damage group id.
    :type damage_group_id_cols: list[str]

    :param hazard_group_id_cols: Columns to be used to generate a hashed hazard group id.
    :type hazard_group_id_cols: list[str]

    :return: GUL inputs dataframe
    :rtype: pandas.DataFrame

    """
    # Get the grouped exposure profile - this describes the financial terms to
    # to be found in the source exposure file, which are for the following
    # FM levels: site coverage (# 1), site pd (# 2), site all (# 3). It also
    # describes the OED hierarchy terms present in the exposure file, namely
    # portfolio num., acc. num., loc. num., and cond. num.

    profile = get_grouped_fm_profile_by_level_and_term_group(exposure_profile=exposure_profile)

    if not profile:
        raise OasisException(
            'Source exposure profile is possibly missing FM term information: '
            'FM term definitions for TIV, limit, deductible, attachment and/or share.'
        )

    # Get the OED hierarchy terms profile - this defines the column names for loc.
    # ID, acc. ID, policy no. and portfolio no., as used in the source exposure
    # and accounts files. This is to ensure that the method never makes hard
    # coded references to the corresponding columns in the source files, as
    # that would mean that changes to these column names in the source files
    # may break the method
    oed_hierarchy = get_oed_hierarchy(exposure_profile=exposure_profile)
    loc_num = oed_hierarchy['locnum']['ProfileElementName']
    acc_num = oed_hierarchy['accnum']['ProfileElementName']
    portfolio_num = oed_hierarchy['portnum']['ProfileElementName']

    # The (site) coverage FM level ID (# 1 in the OED FM levels hierarchy)
    cov_level_id = SUPPORTED_FM_LEVELS['site coverage']['id']

    # Get the TIV column names and corresponding coverage types
    tiv_terms = OrderedDict({v['tiv']['CoverageTypeID']: v['tiv']['ProfileElementName'] for k, v in profile[cov_level_id].items()})
    tiv_cols = list(tiv_terms.values())

    # Get the list of coverage type IDs - financial terms for the coverage
    # level are grouped by coverage type ID in the grouped version of the
    # exposure profile (profile of the financial terms sourced from the
    # source exposure file)
    cov_types = [v['id'] for v in SUPPORTED_COVERAGE_TYPES.values()]

    # Get the FM terms profile (this is a simplfied view of the main grouped
    # profile, containing only information about the financial terms), and
    # the list of OED colum names for the financial terms for the site coverage
    # (# 1 ) FM level
    fm_terms = get_grouped_fm_terms_by_level_and_term_group(grouped_profile_by_level_and_term_group=profile, lowercase=False)
    terms_floats = ['deductible', 'deductible_min', 'deductible_max', 'limit']
    terms_ints = ['ded_code', 'ded_type', 'lim_code', 'lim_type']
    terms = terms_floats + terms_ints
    term_cols_floats = get_fm_terms_oed_columns(
        fm_terms,
        levels=['site coverage'],
        term_group_ids=cov_types,
        terms=terms_floats
    )
    term_cols_ints = get_fm_terms_oed_columns(
        fm_terms,
        levels=['site coverage'],
        term_group_ids=cov_types,
        terms=terms_ints
    )
    term_cols = term_cols_floats + term_cols_ints

    # Create the basic GUL inputs dataframe from merging the exposure and
    # keys dataframes on loc. number/loc. ID; filter out any rows with
    # zeros for TIVs for all coverage types, and replace any nulls in the
    # cond.num. and TIV columns with zeros

    # add default values if missing
    if 'IsAggregate' not in location_df.columns:
        location_df['IsAggregate'] = 0
    else:
        location_df['IsAggregate'].fillna(0, inplace=True)

    # Make sure NumberOfBuildings is there and filled (not mandatory), otherwise assume NumberOfBuildings = 1
    if 'NumberOfBuildings' not in location_df.columns:
        location_df['NumberOfBuildings'] = 1
    else:
        location_df['NumberOfBuildings'].fillna(1, inplace=True)

    # Select only the columns required. This reduces memory use significantly for portfolios
    # that include many OED columns.
    exposure_df_gul_inputs_cols = ['loc_id', portfolio_num, acc_num, loc_num, 'NumberOfBuildings', 'IsAggregate'] + term_cols + tiv_cols
    if SOURCE_IDX['loc'] in location_df:
        exposure_df_gul_inputs_cols += [SOURCE_IDX['loc']]

    # it is assumed that correlations are False for now, correlations for group ID hashing are assessed later on in
    # the process to re-hash the group ID with the correlation "peril_correlation_group" column name. This is because
    # the correlations is achieved later in the process leading to a chicken and egg problem
    # group_id_cols = process_group_id_cols(group_id_cols=group_id_cols,
    #                                       exposure_df_columns=list(exposure_df.columns),
    #                                       has_correlation_groups=False)

    # set damage_group_id_cols
    if not damage_group_id_cols:
        # damage_group_id_cols is None or an empty list
        damage_group_id_cols = DAMAGE_GROUP_ID_COLS
    else:
        # remove any duplicate column names used to assign group_id
        damage_group_id_cols = list(set(damage_group_id_cols))

    # only add damage group col if not an internal oasis col or if not present already in exposure_df_gul_inputs_cols
    for col in damage_group_id_cols:
        if col in VALID_OASIS_GROUP_COLS:
            pass
        elif col not in exposure_df_gul_inputs_cols:
            exposure_df_gul_inputs_cols.append(col)

    # set hazard_group_id_cols
    if not hazard_group_id_cols:
        # hazard_group_id_cols is None or an empty list
        hazard_group_id_cols = HAZARD_GROUP_ID_COLS
    else:
        # remove any duplicate column names used to assign group_id
        hazard_group_id_cols = list(set(hazard_group_id_cols))

    # only add hazard group col if not an internal oasis col or if not present already in exposure_df_gul_inputs_cols
    for col in hazard_group_id_cols:
        if col in VALID_OASIS_GROUP_COLS:
            pass
        elif col not in exposure_df_gul_inputs_cols:
            exposure_df_gul_inputs_cols.append(col)

    # Check if correlation group field is used to drive damage group id
    # and test that it's present and poulated with integers

    correlation_group_id = CORRELATION_GROUP_ID
    correlation_field = correlation_group_id[0]
    correlation_check = False
    if damage_group_id_cols == correlation_group_id:
        if correlation_field in location_df.columns:
            if location_df[correlation_field].astype('uint32').isnull().sum() == 0:
                correlation_check = True

    query_nonzero_tiv = " | ".join(f"({tiv_col} != 0)" for tiv_col in tiv_cols)
    location_df.loc[:, tiv_cols] = location_df.loc[:, tiv_cols].fillna(0.0)
    location_df.query(query_nonzero_tiv, inplace=True, engine='numexpr')

    gul_inputs_df = location_df[list(set(exposure_df_gul_inputs_cols).intersection(location_df.columns))]
    gul_inputs_df.drop_duplicates('loc_id', inplace=True, ignore_index=True)

    # Rename the main keys dataframe columns - this is due to the fact that the
    # keys file headers use camel case, and don't use underscored names, which
    # is the convention used for the GUL and IL inputs dataframes in the MDK
    keys_df.rename(
        columns={
            'locid': 'loc_id' if 'loc_id' not in keys_df else 'locid',
            'perilid': 'peril_id',
            'coveragetypeid': 'coverage_type_id',
            'areaperilid': 'areaperil_id',
            'vulnerabilityid': 'vulnerability_id',
            'amplificationid': 'amplification_id',
            'modeldata': 'model_data'
        },
        inplace=True,
        copy=False  # Pandas copies column data by default on rename
    )

    # If the keys file relates to a complex/custom model then look for a
    # ``modeldata`` column in the keys file, and ignore the area peril
    # and vulnerability ID columns
    if 'model_data' in keys_df:
        keys_df['areaperil_id'] = keys_df['vulnerability_id'] = -1

    gul_inputs_df = merge_dataframes(
        gul_inputs_df,
        keys_df,
        join_on='loc_id',
        how='inner'
    )

    if gul_inputs_df.empty:
        raise OasisException(
            'Inner merge of the exposure file dataframe '
            'and the keys file dataframe on loc. number/loc. ID '
            'is empty - '
            'please check that the loc. number and loc. ID columns '
            'in the exposure and keys files respectively have a non-empty '
            'intersection'
        )

    # Free memory after merge, before memory-intensive restructuring of data
    del keys_df

    # make query to retain only rows with positive TIV for each coverage type, e.g.: (coverage_type_id == 1 and BuildingsTIV > 0.0) or (...)
    positive_TIV_query = " or ".join(
        map(lambda cov_type: f"(coverage_type_id == {cov_type} and {tiv_terms[cov_type]} > 0.0)", gul_inputs_df.coverage_type_id.unique()))

    gul_inputs_df[tiv_terms.values()].fillna(0, inplace=True)  # convert null T&C values to 0
    gul_inputs_df.query(positive_TIV_query, inplace=True)  # remove rows with TIV=null or TIV=0

    if gul_inputs_df.empty:
        raise OasisException('Empty gul_inputs_df dataframe after dropping rows with zero tiv: please check the exposure input files')

    # prepare column mappings for all coverage types
    cols_by_cov_type = {}
    for cov_type in gul_inputs_df.coverage_type_id.unique():
        tiv_col = tiv_terms[cov_type]
        other_cov_types = [v['id'] for v in SUPPORTED_COVERAGE_TYPES.values() if v['id'] != cov_type]
        other_cov_type_term_cols = (
            [v for k, v in tiv_terms.items() if k != cov_type] +
            get_fm_terms_oed_columns(fm_terms=fm_terms, levels=['site coverage'], term_group_ids=other_cov_types, terms=terms)
        )
        is_bi_coverage = cov_type == SUPPORTED_COVERAGE_TYPES['bi']['id']  # store for cov_type
        cov_type_terms = [t for t in terms if fm_terms[cov_level_id][cov_type].get(t)]
        cov_type_term_cols = get_fm_terms_oed_columns(fm_terms, levels=['site coverage'], term_group_ids=[cov_type], terms=cov_type_terms)
        column_mapping_dict = {
            generic_col: cov_col
            for generic_col, cov_col in zip(cov_type_term_cols, cov_type_terms) if generic_col in gul_inputs_df.columns
        }
        column_mapping_dict[tiv_col] = 'tiv'

        cols_by_cov_type[cov_type] = {
            'to_drop': other_cov_types + other_cov_type_term_cols,
            'is_bi_coverage': is_bi_coverage,
            'column_mapping_dict': column_mapping_dict
        }

    # coverage unpacking and disaggregation loop:
    #  - one row representing N coverages is being transformed to N rows, one per coverage.
    #  - if NumberOfBuildings > 1, on top of unpacking the coverage, it performs the disaggregation of the items
    #    by repeating the rows `NumberOfBuildings` times and assigning to each row a unique `disagg_id`` number,
    #    useful for generating `item_id` later.
    #  - group the rows in the GUL inputs table by coverage type
    #  - set the IL terms (and BI coverage boolean) in each group and update the corresponding frame section in the GUL inputs table
    gul_inputs_reformatted_chunks = []
    terms_found = set()
    for (number_of_buildings, cov_type), cov_type_group in gul_inputs_df.groupby(by=['NumberOfBuildings', 'coverage_type_id'], sort=True):
        # drop columns corresponding to other cov types
        cov_type_group.drop(
            columns=cols_by_cov_type[cov_type]['to_drop'],
            errors="ignore",  # Ignore if any of these cols don't exist
            inplace=True
        )

        # check if coverage type is "bi"
        cov_type_group['is_bi_coverage'] = cols_by_cov_type[cov_type]['is_bi_coverage']

        cov_type_group.rename(columns=cols_by_cov_type[cov_type]['column_mapping_dict'], inplace=True, copy=False)
        cov_type_group['coverage_type_id'] = cov_type
        terms_found.update(cols_by_cov_type[cov_type]['column_mapping_dict'].values())

        # split TIV
        cov_type_group['tiv'] /= max(number_of_buildings, 1)

        # if NumberOfBuildings == 0: still add one entry
        disagg_df_chunk = []
        for building_id in range(1, max(number_of_buildings, 1) + 1):
            disagg_df_chunk.append(cov_type_group.copy().assign(building_id=building_id))

        gul_inputs_reformatted_chunks.append(pd.concat(disagg_df_chunk))

    # concatenate all the unpacked chunks. Sort by index to preserve `item_id` order as in the original code
    gul_inputs_df = (
        pd.concat(gul_inputs_reformatted_chunks)
        .fillna(value={c: 0 for c in terms_found})
        .sort_index()
        .reset_index(drop=True)
        .fillna(value={c: 0 for c in set(gul_inputs_df.columns).intersection(set(term_cols_ints + terms_ints))})
    )

    # set default values and data types for BI coverage boolean, TIV, deductibles and limit
    dtypes = {
        **{t: 'uint8' for t in term_cols_ints + terms_ints},
        **{'is_bi_coverage': 'bool'}
    }
    gul_inputs_df = set_dataframe_column_dtypes(gul_inputs_df, dtypes)

    # set 'disagg_id', `item_id` and `coverage_id`
    gul_inputs_df['item_id'] = factorize_ndarray(
        gul_inputs_df.loc[:, ['loc_id', 'peril_id', 'coverage_type_id', 'building_id']].values, col_idxs=range(4))[0]
    gul_inputs_df['coverage_id'] = factorize_ndarray(gul_inputs_df.loc[:, ['loc_id', 'building_id', 'coverage_type_id']].values, col_idxs=range(3))[0]

    # set default data types
    gul_inputs_df = set_dataframe_column_dtypes(gul_inputs_df, {'item_id': 'int32', 'coverage_id': 'int32'})

    # Set the group ID
    # If the group id is set according to the correlation group field then map this field
    # directly, otherwise create an index of the group id fields

    # keep group_id consistance by lower casing column names and sorting
    damage_group_id_cols_map = {c: c.lower() for c in sorted(damage_group_id_cols)}        # mapping from PascalCase -> 'lower_case'
    hazard_group_id_cols_map = {c: c.lower() for c in sorted(hazard_group_id_cols)}        # mapping from PascalCase -> 'lower_case'

    if correlation_check is True:
        gul_inputs_df['group_id'] = gul_inputs_df[correlation_group_id]

    if correlations:
        # do merge with peril correlation df
        gul_inputs_df = gul_inputs_df.merge(peril_correlation_group_df, left_on='peril_id', right_on='id').reset_index()

    gul_inputs_df["group_id"] = (
        pd.util.hash_pandas_object(
            gul_inputs_df.rename(columns=damage_group_id_cols_map)[sorted(list(damage_group_id_cols_map.values()))], index=False).to_numpy() >> 33
    ).astype('uint32')

    gul_inputs_df["hazard_group_id"] = (
        pd.util.hash_pandas_object(
            gul_inputs_df.rename(columns=hazard_group_id_cols_map)[sorted(list(hazard_group_id_cols_map.values()))], index=False).to_numpy() >> 33
    ).astype('uint32')

    # Select only required columns
    # Order here matches test output expectations
    keyscols = ['peril_id', 'coverage_type_id', 'tiv', 'areaperil_id', 'vulnerability_id']
    if 'amplification_id' in gul_inputs_df.columns:
        keyscols += ['amplification_id']
    usecols = (
        ['loc_id', portfolio_num, acc_num, loc_num] +
        ([SOURCE_IDX['loc']] if SOURCE_IDX['loc'] in gul_inputs_df else []) +
        keyscols +
        terms +
        (['model_data'] if 'model_data' in gul_inputs_df else []) +
        # disagg_id is needed for fm_summary_map
        ['is_bi_coverage', 'group_id', 'coverage_id', 'item_id', 'status', 'building_id', 'NumberOfBuildings', 'IsAggregate', ] +
        (["peril_correlation_group", "damage_correlation_value", 'hazard_group_id', "hazard_correlation_value"] if correlations is True else [])
    )

    usecols = [col for col in usecols if col in gul_inputs_df]

    gul_inputs_df = (
        gul_inputs_df
        [usecols]
        .drop_duplicates(subset='item_id')
        .sort_values("item_id")
        .reindex(columns=list(usecols))
    )

    return gul_inputs_df


@oasis_log
def write_complex_items_file(gul_inputs_df, complex_items_fp, chunksize=100000):
    """
    Writes a complex model items file.

    :param gul_inputs_df: GUL inputs dataframe
    :type gul_inputs_df: pandas.DataFrame

    :param complex_items_fp: Complex/custom model items file path
    :type complex_items_fp: str

    :return: Complex/custom model items file path
    :rtype: str
    """
    try:
        gul_inputs_df.loc[:, ['item_id', 'coverage_id', 'model_data', 'group_id']].drop_duplicates().to_csv(
            path_or_buf=complex_items_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(complex_items_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException("Exception raised in 'write_complex_items_file'", e)


@oasis_log
def write_amplifications_file(gul_inputs_df, amplifications_fp, chunksize=100000):
    """
    Writes an amplifications file. This is the mapping between item IDs and
    amplifications IDs.

    :param gul_inputs_df: GUL inputs dataframe
    :type gul_inputs_df: pandas.DataFrame

    :param amplifications_fp: amplifications file path
    :type amplifications_fp: str

    :return: amplifications file path
    :rtype: str
    """
    try:
        gul_inputs_df.loc[:, ['item_id', 'amplification_id']].drop_duplicates().to_csv(
            path_or_buf=amplifications_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(amplifications_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException("Exception raise in 'write_amplifications_file'", e)

    return amplifications_fp


@oasis_log
def write_items_file(gul_inputs_df, items_fp, chunksize=100000):
    """
    Writes an items file.

    :param gul_inputs_df: GUL inputs dataframe
    :type gul_inputs_df: pandas.DataFrame

    :param items_fp: Items file path
    :type items_fp: str

    :return: Items file path
    :rtype: str
    """
    try:
        gul_inputs_df.loc[:, ['item_id', 'coverage_id', 'areaperil_id', 'vulnerability_id', 'group_id']].drop_duplicates().to_csv(
            path_or_buf=items_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(items_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException("Exception raised in 'write_items_file'", e)

    return items_fp


@oasis_log
def write_coverages_file(gul_inputs_df, coverages_fp, chunksize=100000):
    """
    Writes a coverages file.

    :param gul_inputs_df: GUL inputs dataframe
    :type gul_inputs_df: pandas.DataFrame

    :param coverages_fp: Coverages file path
    :type coverages_fp: str

    :return: Coverages file path
    :rtype: str
    """
    try:
        gul_inputs_df.loc[:, ['coverage_id', 'tiv']].drop_duplicates().to_csv(
            path_or_buf=coverages_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(coverages_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException("Exception raised in 'write_coverages_file'", e)

    return coverages_fp


@oasis_log
def write_gul_input_files(
    gul_inputs_df,
    target_dir,
    correlations_df,
    output_dir,
    oasis_files_prefixes=copy.deepcopy(OASIS_FILES_PREFIXES['gul']),
    chunksize=(2 * 10 ** 5),
):
    """
    Writes the standard Oasis GUL input files to a target directory, using a
    pre-generated dataframe of GUL input items. The files written are
    ::

        items.csv
        coverages.csv

    and optionally a complex items file in case of a complex/custom model.

    :param gul_inputs_df: GUL inputs dataframe
    :type gul_inputs_df: pandas.DataFrame

    :param target_dir: Target directory in which to write the files
    :type target_dir: str

    :param oasis_files_prefixes: Oasis GUL input file name prefixes
    :param oasis_files_prefixes: dict

    :param chunksize: The chunk size to use when writing out the
                      input files
    :type chunksize: int

    :return: GUL input files dict
    :rtype: dict
    """
    # Clean the target directory path
    target_dir = as_path(target_dir, 'Target IL input files directory', is_dir=True, preexists=False)

    if correlations_df is None:
        correlations_df = pd.DataFrame(columns=CorrelationsData.COLUMNS)

    # write the correlations to a binary file
    correlation_data_handle = CorrelationsData(data=correlations_df)
    correlation_data_handle.to_bin(file_path=f"{output_dir}/correlations.bin")
    correlation_data_handle.to_csv(file_path=f"{output_dir}/correlations.csv")

    # Set chunk size for writing the CSV files - default is the minimum of 100K
    # or the GUL inputs frame size
    chunksize = chunksize or min(chunksize, len(gul_inputs_df))

    # If no complex model data present then remove the corresponding file
    # name from the files prefixes dict, which is used for writing the
    # GUl input files
    if 'model_data' not in gul_inputs_df:
        oasis_files_prefixes.pop('complex_items', None)

    # If no amplification IDs then remove corresponding file name from files
    # prefixes dict
    if 'amplification_id' not in gul_inputs_df:
        oasis_files_prefixes.pop('amplifications', None)

    # A dict of GUL input file names and file paths
    gul_input_files = {
        fn: os.path.join(target_dir, '{}.csv'.format(oasis_files_prefixes[fn]))
        for fn in oasis_files_prefixes
    }
    this_module = sys.modules[__name__]
    # Write the files serially
    for fn in gul_input_files:
        getattr(this_module, 'write_{}_file'.format(fn))(gul_inputs_df.copy(deep=True), gul_input_files[fn], chunksize)

    return gul_input_files



"""
This file contains the utilities for generating random numbers in gulpy.

"""

import logging
from math import sqrt

import numpy as np
from numba import njit
from scipy.stats import norm

logger = logging.getLogger(__name__)


GROUP_ID_HASH_CODE = np.int64(1543270363)
EVENT_ID_HASH_CODE = np.int64(1943272559)
HASH_MOD_CODE = np.int64(2147483648)

HAZARD_GROUP_ID_HASH_CODE = np.int64(1143271949)
HAZARD_EVENT_ID_HASH_CODE = np.int64(1243274353)
HAZARD_HASH_MOD_CODE = np.int64(1957483729)


@njit(cache=True, fastmath=True)
def generate_hash(group_id, event_id, base_seed=0):
    """Generate hash for a given `group_id`, `event_id` pair for the vulnerability pdf.

    Args:
        group_id (int): group id.
        event_id (int]): event id.
        base_seed (int, optional): base random seed. Defaults to 0.

    Returns:
        int64: hash
    """
    hash = (base_seed + (group_id * GROUP_ID_HASH_CODE) % HASH_MOD_CODE +
            (event_id * EVENT_ID_HASH_CODE) % HASH_MOD_CODE) % HASH_MOD_CODE

    return hash


@njit(cache=True, fastmath=True)
def generate_hash_hazard(hazard_group_id, event_id, base_seed=0):
    """Generate hash for a given `hazard_group_id`, `event_id` pair for the hazard pdf.

    Args:
        hazard_group_id (int): group id.
        event_id (int]): event id.
        base_seed (int, optional): base random seed. Defaults to 0.

    Returns:
        int64: hash
    """
    hash = (base_seed + (hazard_group_id * HAZARD_GROUP_ID_HASH_CODE) % HAZARD_HASH_MOD_CODE +
            (event_id * HAZARD_EVENT_ID_HASH_CODE) % HAZARD_HASH_MOD_CODE) % HAZARD_HASH_MOD_CODE

    return hash


def get_random_generator(random_generator):
    """Get the random generator function.

    Args:
        random_generator (int): random generator function id.

    Returns:
        The random generator function.

    """
    # define random generator function
    if random_generator == 0:
        logger.info("Random generator: MersenneTwister")
        return random_MersenneTwister

    elif random_generator == 1:
        logger.info("Random generator: Latin Hypercube")
        return random_LatinHypercube

    else:
        raise ValueError(f"No random generator exists for random_generator={random_generator}.")


EVENT_ID_HASH_CODE = np.int64(1943_272_559)
PERIL_CORRELATION_GROUP_HASH = np.int64(1836311903)
HASH_MOD_CODE = np.int64(2147483648)


@njit(cache=True, fastmath=True)
def generate_correlated_hash_vector(unique_peril_correlation_groups, event_id, base_seed=0):
    """Generate hashes for all peril correlation groups for a given `event_id`.

    Args:
        unique_peril_correlation_groups (List[int]): list of the unique peril correlation groups.
        event_id (int): event id.
        base_seed (int, optional): base random seed. Defaults to 0.

    Returns:
        List[int64]: hashes
    """
    Nperil_correlation_groups = np.max(unique_peril_correlation_groups)
    correlated_hashes = np.zeros(Nperil_correlation_groups + 1, dtype='int64')
    correlated_hashes[0] = 0

    unique_peril_index = 0
    unique_peril_len = unique_peril_correlation_groups.shape[0]
    for i in range(1, Nperil_correlation_groups + 1):
        if unique_peril_correlation_groups[unique_peril_index] == i:
            correlated_hashes[i] = (
                base_seed +
                (unique_peril_correlation_groups[unique_peril_index] * PERIL_CORRELATION_GROUP_HASH) % HASH_MOD_CODE +
                (event_id * EVENT_ID_HASH_CODE) % HASH_MOD_CODE
            ) % HASH_MOD_CODE
            unique_peril_index += 1
            if unique_peril_index == unique_peril_len:
                break

    return correlated_hashes


def compute_norm_inv_cdf_lookup(cdf_min, cdf_max, N):
    return norm.ppf(np.linspace(cdf_min, cdf_max, N))


def compute_norm_cdf_lookup(x_min, x_max, N):
    return norm.cdf(np.linspace(x_min, x_max, N))


@njit(cache=True, fastmath=True)
def get_norm_cdf_cell_nb(x, x_min, x_max, N):
    return int((x - x_min) * (N - 1) // (x_max - x_min))


@njit(cache=True, fastmath=True)
def get_corr_rval(x_unif, y_unif, rho, x_min, x_max, N, norm_inv_cdf, cdf_min,
                  cdf_max, norm_cdf, Nsamples, z_unif):

    sqrt_rho = sqrt(rho)
    sqrt_1_minus_rho = sqrt(1. - rho)

    for i in range(Nsamples):
        x_norm = norm_inv_cdf[get_norm_cdf_cell_nb(x_unif[i], x_min, x_max, N)]
        y_norm = norm_inv_cdf[get_norm_cdf_cell_nb(y_unif[i], x_min, x_max, N)]
        z_norm = sqrt_rho * x_norm + sqrt_1_minus_rho * y_norm

        z_unif[i] = norm_cdf[get_norm_cdf_cell_nb(z_norm, cdf_min, cdf_max, N)]


@njit(cache=True, fastmath=True)
def random_MersenneTwister(seeds, n, skip_seeds=0):
    """Generate random numbers using the default Mersenne Twister algorithm.

    Args:
        seeds (List[int64]): List of seeds.
        n (int): number of random samples to generate for each seed.
        skip_seeds (int): number of seeds to skip starting from the beginning
          of the `seeds` array. For skipped seeds no random numbers are generated
          and the output rndms will contain zeros at their corresponding row.
          Default is 0, i.e. no seeds are skipped.

    Returns:
        rndms (array[float]): 2-d array of shape (number of seeds, n) 
          containing the random values generated for each seed.
        rndms_idx (Dict[int64, int]): mapping between `seed` and the 
          row in rndms that stores the corresponding random values.
    """
    Nseeds = len(seeds)
    rndms = np.zeros((Nseeds, n), dtype='float64')

    for seed_i in range(skip_seeds, Nseeds, 1):
        # set the seed
        np.random.seed(seeds[seed_i])

        # draw the random numbers
        for j in range(n):
            # by default in numba this should be Mersenne-Twister
            rndms[seed_i, j] = np.random.uniform(0., 1.)

    return rndms


@njit(cache=True, fastmath=True)
def random_LatinHypercube(seeds, n, skip_seeds=0):
    """Generate random numbers using the Latin Hypercube algorithm.

    Args:
        seeds (List[int64]): List of seeds.
        n (int): number of random samples to generate for each seed.

    Returns:
        rndms (array[float]): 2-d array of shape (number of seeds, n) 
          containing the random values generated for each seed.
        rndms_idx (Dict[int64, int]): mapping between `seed` and the 
          row in rndms that stores the corresponding random values.
        skip_seeds (int): number of seeds to skip starting from the beginning
          of the `seeds` array. For skipped seeds no random numbers are generated
          and the output rndms will contain zeros at their corresponding row.
          Default is 0, i.e. no seeds are skipped.

    Notes:
        Implementation follows scipy.stats.qmc.LatinHypercube v1.8.0.
        Following scipy notation, here we assume `centered=False` all the times:
        instead of taking `samples=0.5*np.ones(n)`, here we always
        draw uniform random samples in order to initialise `samples`.
    """
    Nseeds = len(seeds)
    rndms = np.zeros((Nseeds, n), dtype='float64')
    # define arrays here and re-use them later
    samples = np.zeros(n, dtype='float64')
    perms = np.zeros(n, dtype='float64')

    for seed_i in range(skip_seeds, Nseeds, 1):
        # set the seed
        np.random.seed(seeds[seed_i])

        # draw the random numbers and re-generate permutations array
        for i in range(n):
            samples[i] = np.random.uniform(0., 1.)
            perms[i] = i + 1

        # in-place shuffle permutations
        np.random.shuffle(perms)

        for j in range(n):
            rndms[seed_i, j] = (perms[j] - samples[j]) / float(n)

    return rndms



__all__ = [
    'GenerateFiles',
    'GenerateDummyModelFiles',
    'GenerateDummyOasisFiles'
]

import io
import json
import os
from pathlib import Path
from typing import List

from oasislmf.computation.base import ComputationStep
from oasislmf.computation.data.dummy_model.generate import (CoveragesFile,
                                                            DamageBinDictFile,
                                                            EventsFile,
                                                            FMPolicyTCFile,
                                                            FMProfileFile,
                                                            FMProgrammeFile,
                                                            FMSummaryXrefFile,
                                                            FMXrefFile,
                                                            FootprintBinFile,
                                                            GULSummaryXrefFile,
                                                            ItemsFile,
                                                            OccurrenceFile,
                                                            RandomFile,
                                                            VulnerabilityFile)
from oasislmf.computation.generate.keys import GenerateKeys
from oasislmf.preparation.correlations import map_data
from oasislmf.preparation.dir_inputs import (create_target_directory,
                                             prepare_input_files_directory)
from oasislmf.preparation.gul_inputs import (get_gul_input_items,
                                             process_group_id_cols,
                                             write_gul_input_files)
from oasislmf.preparation.il_inputs import (get_il_input_items,
                                            get_oed_hierarchy,
                                            write_il_input_files)
from oasislmf.preparation.reinsurance_layer import write_files_for_reinsurance
from oasislmf.preparation.summaries import (get_summary_mapping,
                                            merge_oed_to_mapping,
                                            write_exposure_summary,
                                            write_mapping_file,
                                            write_summary_levels)
from oasislmf.pytools.data_layer.oasis_files.correlations import \
    CorrelationsData
from oasislmf.utils.data import (establish_correlations, get_dataframe,
                                 get_exposure_data, get_json,
                                 get_utctimestamp,
                                 prepare_account_df, prepare_location_df,
                                 prepare_reinsurance_df)
from oasislmf.utils.defaults import (DAMAGE_GROUP_ID_COLS,
                                     HAZARD_GROUP_ID_COLS,
                                     OASIS_FILES_PREFIXES, WRITE_CHUNKSIZE,
                                     get_default_accounts_profile,
                                     get_default_exposure_profile,
                                     get_default_fm_aggregation_profile)
from oasislmf.utils.exceptions import OasisException
from oasislmf.utils.inputs import str2bool
from ods_tools.oed.setting_schema import ModelSettingSchema


class GenerateFiles(ComputationStep):
    """
    Generates the standard Oasis GUL input files + optionally the IL/FM input
    files and the RI input files.
    """
    step_params = [
        # Command line options
        {'name': 'oasis_files_dir', 'flag': '-o', 'is_path': True, 'pre_exist': False,
         'help': 'Path to the directory in which to generate the Oasis files'},
        {'name': 'keys_data_csv', 'flag': '-z', 'is_path': True, 'pre_exist': True, 'help': 'Pre-generated keys CSV file path'},
        {'name': 'keys_errors_csv', 'is_path': True, 'pre_exist': True, 'help': 'Pre-generated keys errors CSV file path'},
        {'name': 'lookup_config_json', 'flag': '-m', 'is_path': True, 'pre_exist': False, 'help': 'Lookup config JSON file path'},
        {'name': 'lookup_data_dir', 'flag': '-k', 'is_path': True, 'pre_exist': True, 'help': 'Model lookup/keys data directory path'},
        {'name': 'lookup_module_path', 'flag': '-l', 'is_path': True, 'pre_exist': False, 'help': 'Model lookup module path'},
        {'name': 'lookup_complex_config_json', 'flag': '-L', 'is_path': True, 'pre_exist': False, 'help': 'Complex lookup config JSON file path'},
        {'name': 'lookup_num_processes', 'type': int, 'default': -1, 'help': 'Number of workers in multiprocess pools'},
        {'name': 'lookup_num_chunks', 'type': int, 'default': -1, 'help': 'Number of chunks to split the location file into for multiprocessing'},
        {'name': 'model_version_csv', 'flag': '-v', 'is_path': True, 'pre_exist': False, 'help': 'Model version CSV file path'},
        {'name': 'model_settings_json', 'flag': '-M', 'is_path': True, 'pre_exist': True, 'help': 'Model settings JSON file path'},
        {'name': 'user_data_dir', 'flag': '-D', 'is_path': True, 'pre_exist': False,
         'help': 'Directory containing additional model data files which varies between analysis runs'},
        {'name': 'profile_loc_json', 'flag': '-e', 'is_path': True, 'pre_exist': True, 'help': 'Source (OED) exposure profile JSON path'},
        {'name': 'profile_acc_json', 'flag': '-b', 'is_path': True, 'pre_exist': True, 'help': 'Source (OED) accounts profile JSON path'},
        {'name': 'profile_fm_agg_json', 'flag': '-g', 'is_path': True, 'pre_exist': True, 'help': 'FM (OED) aggregation profile path'},
        {'name': 'oed_schema_info', 'is_path': True, 'pre_exist': True, 'help': 'path to custom oed_schema'},
        {'name': 'currency_conversion_json', 'is_path': True, 'pre_exist': True, 'help': 'settings to perform currency conversion of oed files'},
        {'name': 'reporting_currency', 'help': 'currency to use in the results reported'},
        {'name': 'oed_location_csv', 'flag': '-x', 'is_path': True, 'pre_exist': True, 'help': 'Source location CSV file path'},
        {'name': 'oed_accounts_csv', 'flag': '-y', 'is_path': True, 'pre_exist': True, 'help': 'Source accounts CSV file path'},
        {'name': 'oed_info_csv', 'flag': '-i', 'is_path': True, 'pre_exist': True, 'help': 'Reinsurance info. CSV file path'},
        {'name': 'oed_scope_csv', 'flag': '-s', 'is_path': True, 'pre_exist': True, 'help': 'Reinsurance scope CSV file path'},
        {'name': 'check_oed', 'type': str2bool, 'const': True, 'nargs': '?', 'default': True, 'help': 'if True check input oed files'},
        {'name': 'disable_summarise_exposure', 'flag': '-S', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'Disables creation of an exposure summary report'},
        {'name': 'damage_group_id_cols', 'flag': '-G', 'nargs': '+', 'help': 'Columns from loc file to set group_id', 'default': DAMAGE_GROUP_ID_COLS},
        {'name': 'hazard_group_id_cols', 'flag': '-H', 'nargs': '+', 'help': 'Columns from loc file to set hazard_group_id', 'default': HAZARD_GROUP_ID_COLS},
        {'name': 'lookup_multiprocessing', 'type': str2bool, 'const': False, 'nargs': '?', 'default': False,
         'help': 'Flag to enable/disable lookup multiprocessing'},

        # Manager only options (pass data directy instead of filepaths)
        {'name': 'lookup_config'},
        {'name': 'lookup_complex_config'},
        {'name': 'write_ri_tree', 'default': False},
        {'name': 'verbose', 'default': False},
        {'name': 'write_chunksize', 'type': int, 'default': WRITE_CHUNKSIZE},
        {'name': 'oasis_files_prefixes', 'default': OASIS_FILES_PREFIXES},
        {'name': 'profile_loc', 'default': get_default_exposure_profile()},
        {'name': 'profile_acc', 'default': get_default_accounts_profile()},
        {'name': 'profile_fm_agg', 'default': get_default_fm_aggregation_profile()},
    ]

    def _get_output_dir(self):
        if self.oasis_files_dir:
            return self.oasis_files_dir
        utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
        return os.path.join(os.getcwd(), 'runs', 'files-{}'.format(utcnow))

    def get_exposure_data_config(self):
        return {
            'location': self.oed_location_csv,
            'account': self.oed_accounts_csv,
            'ri_info': self.oed_info_csv,
            'ri_scope': self.oed_scope_csv,
            'oed_schema_info': self.oed_schema_info,
            'currency_conversion': self.currency_conversion_json,
            'check_oed': self.check_oed,
            'use_field': True
        }

    def run(self):
        self.logger.info('\nProcessing arguments - Creating Oasis Files')

        if not (self.keys_data_csv or self.lookup_config_json or (self.lookup_data_dir and self.model_version_csv and self.lookup_module_path)):
            raise OasisException(
                'No pre-generated keys file provided, and no lookup assets '
                'provided to generate a keys file - if you do not have a '
                'pre-generated keys file then lookup assets must be provided - '
                'for a built-in lookup the lookup config. JSON file path must '
                'be provided, or for custom lookups the keys data path + model '
                'version file path + lookup package path must be provided'
            )

        self.oasis_files_dir = self._get_output_dir()
        exposure_data = get_exposure_data(self, add_internal_col=True)
        self.kwargs['exposure_data'] = exposure_data

        il = bool(exposure_data.account)
        ri = exposure_data.ri_info and exposure_data.ri_scope and il
        self.logger.info('\nGenerating Oasis files (GUL=True, IL={}, RIL={})'.format(il, ri))
        summarise_exposure = not self.disable_summarise_exposure

        # Prepare the target directory and copy the source files, profiles and
        # model version into it
        target_dir = prepare_input_files_directory(
            target_dir=self.oasis_files_dir,
            exposure_data=exposure_data,
            exposure_profile_fp=self.profile_loc_json,
            keys_fp=self.keys_data_csv,
            keys_errors_fp=self.keys_errors_csv,
            lookup_config_fp=self.lookup_config_json,
            model_version_fp=self.model_version_csv,
            complex_lookup_config_fp=self.lookup_complex_config_json,
            accounts_profile_fp=self.profile_acc_json,
            fm_aggregation_profile_fp=self.profile_fm_agg_json,
        )
        # Get the profiles defining the exposure and accounts files, ID related
        # terms in these files, and FM aggregation hierarchy
        location_profile = get_json(src_fp=self.profile_loc_json) if self.profile_loc_json else self.profile_loc
        accounts_profile = get_json(src_fp=self.profile_acc_json) if self.profile_acc_json else self.profile_acc
        oed_hierarchy = get_oed_hierarchy(location_profile, accounts_profile)
        loc_grp = oed_hierarchy['locgrp']['ProfileElementName']

        fm_aggregation_profile = get_json(src_fp=self.profile_fm_agg_json) if self.profile_fm_agg_json else self.profile_fm_agg

        # force fm_agg level keys to type int:
        if any(isinstance(lvl, str) for lvl in fm_aggregation_profile.keys()):
            fm_aggregation_profile = {int(k): v for k, v in fm_aggregation_profile.items()}

        if self.reporting_currency:
            exposure_data.reporting_currency = self.reporting_currency
            exposure_data.save(target_dir, self.reporting_currency, save_config=True)

        exposure_data.location.dataframe = prepare_location_df(exposure_data.location.dataframe)
        location_df = exposure_data.location.dataframe

        if il:
            exposure_data.account.dataframe = prepare_account_df(exposure_data.account.dataframe)
            account_df = exposure_data.account.dataframe
        else:
            account_df = None

        # If a pre-generated keys file path has not been provided,
        # then it is asssumed some model lookup assets have been provided, so
        # as to allow the lookup to be instantiated and called to generated
        # the keys file.
        _keys_fp = _keys_errors_fp = None
        if not self.keys_data_csv:
            _keys_fp = self.kwargs['keys_data_csv'] = os.path.join(target_dir, 'keys.csv')
            _keys_errors_fp = self.kwargs['keys_errors_csv'] = os.path.join(target_dir, 'keys-errors.csv')
            GenerateKeys(**self.kwargs).run()
        else:
            _keys_fp = os.path.join(target_dir, os.path.basename(self.keys_data_csv))
            if self.keys_errors_csv:
                _keys_errors_fp = os.path.join(target_dir, os.path.basename(self.keys_errors_csv))

        # Load keys file  **** WARNING - REFACTOR THIS ****
        dtypes = {
            'locid': 'int64',
            'perilid': 'str',
            'coveragetypeid': 'uint8',
            'areaperilid': 'uint64',
            'vulnerabilityid': 'uint32',
            'modeldata': 'str'
        }
        keys_error_fp = os.path.join(os.path.dirname(_keys_fp), 'keys-errors.csv') if _keys_fp else 'Missing'
        missing_keys_msg = 'No successful lookup results found in the keys file - '
        missing_keys_msg += 'Check the `keys-errors.csv` file for details. \n File path: {}'.format(keys_error_fp)
        keys_df = get_dataframe(
            src_fp=_keys_fp,
            col_dtypes=dtypes,
            empty_data_error_msg=missing_keys_msg,
            memory_map=True
        )
        # ************************************************

        # Columns from loc file to assign group_id
        model_damage_group_fields = []
        model_hazard_group_fields = []
        correlations: bool = False
        model_settings = None

        if self.model_settings_json is not None:
            model_settings = ModelSettingSchema().get(self.model_settings_json)
            correlations = establish_correlations(model_settings=model_settings)
            try:
                model_damage_group_fields = model_settings["data_settings"].get("damage_group_fields")
            except (KeyError, AttributeError, OasisException) as e:
                self.logger.warn(f'WARNING: Failed to load "damage_group_fields", file: {self.model_settings_json}, error: {e}')
            try:
                model_hazard_group_fields = model_settings["data_settings"].get("hazard_group_fields")
            except (KeyError, AttributeError, OasisException) as e:
                self.logger.warn(f'WARNING: Failed to load "hazard_group_fields", file: {self.model_settings_json}, error: {e}')

        # load group columns from model_settings.json if not set in kwargs (CLI)
        if model_damage_group_fields and not self.kwargs.get('group_id_cols'):
            damage_group_id_cols = model_damage_group_fields
        # otherwise load group cols from args
        else:
            damage_group_id_cols = self.damage_group_id_cols

        # load hazard group columns from model_settings.json if not set in kwargs (CLI)
        if model_hazard_group_fields and not self.kwargs.get('hazard_group_id_cols'):
            hazard_group_id_cols = model_hazard_group_fields
        # otherwise load group cols from args
        else:
            hazard_group_id_cols = self.hazard_group_id_cols

        damage_group_id_cols: List[str] = process_group_id_cols(group_id_cols=damage_group_id_cols,
                                                                exposure_df_columns=location_df,
                                                                has_correlation_groups=correlations)

        hazard_group_id_cols: List[str] = process_group_id_cols(group_id_cols=hazard_group_id_cols,
                                                                exposure_df_columns=location_df,
                                                                has_correlation_groups=correlations)

        gul_inputs_df = get_gul_input_items(
            location_df,
            keys_df,
            peril_correlation_group_df=map_data(data=model_settings),
            correlations=correlations,
            exposure_profile=location_profile,
            damage_group_id_cols=damage_group_id_cols,
            hazard_group_id_cols=hazard_group_id_cols,
        )

        # If not in det. loss gen. scenario, write exposure summary file
        if summarise_exposure:
            write_exposure_summary(
                target_dir,
                location_df,
                keys_fp=_keys_fp,
                keys_errors_fp=_keys_errors_fp,
                exposure_profile=location_profile,
            )

        # If exposure summary set, write valid columns for summary levels to file
        if summarise_exposure:
            write_summary_levels(location_df, account_df, exposure_data, target_dir)

        # Write the GUL input files
        files_prefixes = self.oasis_files_prefixes

        gul_input_files = write_gul_input_files(
            gul_inputs_df,
            target_dir,
            correlations_df=gul_inputs_df[CorrelationsData.COLUMNS] if correlations is True else None,
            output_dir=self._get_output_dir(),
            oasis_files_prefixes=files_prefixes['gul'],
            chunksize=self.write_chunksize,
        )
        gul_summary_mapping = get_summary_mapping(gul_inputs_df, oed_hierarchy)
        write_mapping_file(gul_summary_mapping, target_dir)

        # If no source accounts file path has been provided assume that IL
        # input files, and therefore also RI input files, are not needed
        if not il:
            # Write `summary_map.csv` for GUL only
            self.logger.info('\nOasis files generated: {}'.format(json.dumps(gul_input_files, indent=4)))
            return gul_input_files

        # Get the IL input items
        il_inputs_df = get_il_input_items(
            exposure_df=exposure_data.location.dataframe,
            gul_inputs_df=gul_inputs_df,
            accounts_df=exposure_data.account.dataframe,
            exposure_profile=location_profile,
            accounts_profile=accounts_profile,
            fm_aggregation_profile=fm_aggregation_profile
        )

        # Write the IL/FM input files
        il_input_files = write_il_input_files(
            il_inputs_df,
            target_dir,
            oasis_files_prefixes=files_prefixes['il'],
            chunksize=self.write_chunksize
        )

        fm_summary_mapping = get_summary_mapping(il_inputs_df, oed_hierarchy, is_fm_summary=True)
        write_mapping_file(fm_summary_mapping, target_dir, is_fm_summary=True)

        # Combine the GUL and IL input file paths into a single dict (for convenience)
        oasis_files = {**gul_input_files, **il_input_files}

        # If no RI input file paths (info. and scope) have been provided then
        # no RI input files are needed, just return the GUL and IL Oasis files
        if not ri:
            self.logger.info('\nOasis files generated: {}'.format(json.dumps(oasis_files, indent=4)))
            return oasis_files

        exposure_data.ri_info.dataframe, exposure_data.ri_scope.dataframe = prepare_reinsurance_df(exposure_data.ri_info.dataframe,
                                                                                                   exposure_data.ri_scope.dataframe)

        # Write the RI input files, and write the returned RI layer info. as a
        # file, which can be reused by the model runner (in the model execution
        # stage) to set the number of RI iterations
        xref_descriptions_df = merge_oed_to_mapping(
            fm_summary_mapping,
            exposure_data.location.dataframe,
            oed_column_set=[loc_grp],
            defaults={loc_grp: 1}
        ).sort_values(by='agg_id')

        self.kwargs['oed_info_csv'] = exposure_data.ri_info

        ri_layers = write_files_for_reinsurance(
            gul_inputs_df,
            xref_descriptions_df,
            exposure_data.ri_info.dataframe,
            exposure_data.ri_scope.dataframe,
            oasis_files['fm_xref'],
            target_dir,
            self.write_ri_tree
        )

        with io.open(os.path.join(target_dir, 'ri_layers.json'), 'w', encoding='utf-8') as f:
            f.write(json.dumps(ri_layers, ensure_ascii=False, indent=4))
            oasis_files['ri_layers'] = os.path.abspath(f.name)
            for layer, layer_info in ri_layers.items():
                oasis_files['RI_{}'.format(layer)] = layer_info['directory']

        self.logger.info('\nOasis files generated: {}'.format(json.dumps(oasis_files, indent=4)))

        return oasis_files


class GenerateDummyModelFiles(ComputationStep):
    """
    Generates dummy model files.
    """

    # Command line options
    step_params = [
        {'name': 'target_dir', 'flag': '-o', 'is_path': True, 'pre_exist': False,
         'help': 'Path to the directory in which to generate the Model files'},
        {'name': 'num_vulnerabilities', 'flag': '-v', 'required': True, 'type': int, 'help': 'Number of vulnerabilities'},
        {'name': 'num_intensity_bins', 'flag': '-i', 'required': True, 'type': int, 'help': 'Number of intensity bins'},
        {'name': 'num_damage_bins', 'flag': '-d', 'required': True, 'type': int, 'help': 'Number of damage bins'},
        {'name': 'vulnerability_sparseness', 'flag': '-s', 'required': False, 'type': float, 'default': 1.0,
         'help': 'Percentage of bins normalised to range [0,1] impacted for a vulnerability at an intensity level'},
        {'name': 'num_events', 'flag': '-e', 'required': True, 'type': int, 'help': 'Number of events'},
        {'name': 'num_areaperils', 'flag': '-a', 'required': True, 'type': int, 'help': 'Number of areaperils'},
        {'name': 'areaperils_per_event', 'flag': '-A', 'required': False, 'type': int, 'default': None,
         'help': 'Number of areaperils impacted per event'},
        {'name': 'intensity_sparseness', 'flag': '-S', 'required': False, 'type': float, 'default': 1.0,
         'help': 'Percentage of bins normalised to range [0,1] impacted for an event and areaperil'},
        {'name': 'no_intensity_uncertainty', 'flag': '-u', 'required': False,
         'default': False, 'action': 'store_true', 'help': 'No intensity uncertainty flag'},
        {'name': 'num_periods', 'flag': '-p', 'required': True, 'type': int, 'help': 'Number of periods'},
        {'name': 'periods_per_event_mean', 'flag': '-P', 'required': False, 'type': int, 'default': 1,
         'help': 'Mean of truncated normal distribution sampled to determine number of periods per event'},
        {'name': 'periods_per_event_stddev', 'flag': '-Q', 'required': False, 'type': float, 'default': 0.0,
         'help': 'Standard deviation of truncated normal distribution sampled to determine number of periods per event'},
        {'name': 'num_randoms', 'flag': '-r', 'required': False, 'type': int, 'default': 0, 'help': 'Number of random numbers'},
        {'name': 'random_seed', 'flag': '-R', 'required': False, 'type': int, 'default': -
         1, 'help': 'Random seed (-1 for 1234 (default), 0 for current system time'}
    ]

    def _validate_input_arguments(self):
        if self.vulnerability_sparseness > 1.0 or self.vulnerability_sparseness < 0.0:
            raise OasisException('Invalid value for --vulnerability-sparseness')
        if self.intensity_sparseness > 1.0 or self.intensity_sparseness < 0.0:
            raise OasisException('Invlid value for --intensity-sparseness')
        if not self.areaperils_per_event:
            self.areaperils_per_event = self.num_areaperils
        if self.areaperils_per_event > self.num_areaperils:
            raise OasisException('Number of areaperils per event exceeds total number of areaperils')
        if self.random_seed < -1:
            raise OasisException('Invalid random seed')

    def _create_target_directory(self, label):
        utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
        if not self.target_dir:
            self.target_dir = os.path.join(os.getcwd(), 'runs', f'test-{label}-{utcnow}')

        self.target_dir = create_target_directory(
            self.target_dir, 'target test model files directory'
        )

    def _prepare_run_directory(self):
        self.input_dir = os.path.join(self.target_dir, 'input')
        self.static_dir = os.path.join(self.target_dir, 'static')
        directories = [
            self.input_dir, self.static_dir
        ]
        for directory in directories:
            if not os.path.exists(directory):
                Path(directory).mkdir(parents=True, exist_ok=True)

    def _set_footprint_files_inputs(self):
        self.footprint_files_inputs = {
            'num_events': self.num_events,
            'num_areaperils': self.num_areaperils,
            'areaperils_per_event': self.areaperils_per_event,
            'num_intensity_bins': self.num_intensity_bins,
            'intensity_sparseness': self.intensity_sparseness,
            'no_intensity_uncertainty': self.no_intensity_uncertainty,
            'directory': self.static_dir
        }

    def _set_periods_per_event_parameters(self):
        self.periods_per_event_parameters = {
            'mean': self.periods_per_event_mean,
            'stddev': self.periods_per_event_stddev
        }

    def _get_model_file_objects(self):

        # vulnerability.bin, events.bin, footprint.bin, footprint.idx,
        # damage_bin_dict.bin and occurrence.bin
        self._set_footprint_files_inputs()
        self._set_periods_per_event_parameters()
        self.model_files = [
            VulnerabilityFile(
                self.num_vulnerabilities, self.num_intensity_bins,
                self.num_damage_bins, self.vulnerability_sparseness,
                self.random_seed, self.static_dir
            ),
            EventsFile(self.num_events, self.input_dir),
            FootprintBinFile(
                **self.footprint_files_inputs, random_seed=self.random_seed
            ),
            DamageBinDictFile(self.num_damage_bins, self.static_dir),
            OccurrenceFile(
                self.num_events, self.num_periods, self.random_seed,
                self.input_dir, **self.periods_per_event_parameters
            )
        ]
        if self.num_randoms > 0:
            self.model_files += [
                RandomFile(self.num_randoms, self.random_seed, self.static_dir)
            ]

    def run(self):
        self.logger.info('\nProcessing arguments - Creating Dummy Model Files')

        self._validate_input_arguments()
        self._create_target_directory(label='files')
        self._prepare_run_directory()
        self._get_model_file_objects()

        for model_file in self.model_files:
            self.logger.info(f'Writing {model_file.file_name}')
            model_file.write_file()

        self.logger.info(f'\nDummy Model files generated in {self.target_dir}')


class GenerateDummyOasisFiles(GenerateDummyModelFiles):
    """
    Generates dummy model and Oasis GUL input files + optionally the IL/FM
    input files.
    """

    step_params = [
        {'name': 'num_locations', 'flag': '-l', 'required': True, 'type': int, 'help': 'Number of locations'},
        {'name': 'coverages_per_location', 'flag': '-c', 'required': True, 'type': int, 'help': 'Number of coverage types per location'},
        {'name': 'num_layers', 'flag': '-L', 'required': False, 'type': int, 'default': 1, 'help': 'Number of layers'}
    ]
    chained_commands = [GenerateDummyModelFiles]

    def _validate_input_arguments(self):
        super()._validate_input_arguments()
        if self.coverages_per_location > 4 or self.coverages_per_location < 1:
            raise OasisException('Number of supported coverage types is 1 to 4')

    def _get_gul_file_objects(self):

        # coverages.bin, items.bin and gulsummaryxref.bin
        self.gul_files = [
            CoveragesFile(
                self.num_locations, self.coverages_per_location,
                self.random_seed, self.input_dir
            ),
            ItemsFile(
                self.num_locations, self.coverages_per_location,
                self.num_areaperils, self.num_vulnerabilities,
                self.random_seed, self.input_dir
            ),
            GULSummaryXrefFile(
                self.num_locations, self.coverages_per_location, self.input_dir
            )
        ]

    def _get_fm_file_objects(self):

        # fm_programme.bin, fm_policytc.bin, fm_profile.bin, fm_xref.bin and
        # fmsummaryxref.bin
        self.fm_files = [
            FMProgrammeFile(
                self.num_locations, self.coverages_per_location, self.input_dir
            ),
            FMPolicyTCFile(
                self.num_locations, self.coverages_per_location,
                self.num_layers, self.input_dir
            ),
            FMProfileFile(self.num_layers, self.input_dir),
            FMXrefFile(
                self.num_locations, self.coverages_per_location,
                self.num_layers, self.input_dir
            ),
            FMSummaryXrefFile(
                self.num_locations, self.coverages_per_location,
                self.num_layers, self.input_dir
            )
        ]

    def run(self):
        self.logger.info('\nProcessing arguments - Creating Model & Test Oasis Files')

        self._validate_input_arguments()
        self._create_target_directory(label='files')
        self._prepare_run_directory()
        self._get_model_file_objects()
        self._get_gul_file_objects()
        self._get_fm_file_objects()

        output_files = self.model_files + self.gul_files + self.fm_files
        for output_file in output_files:
            self.logger.info(f'Writing {output_file.file_name}')
            output_file.write_file()

        self.logger.info(f'\nDummy Model and Oasis files generated in {self.target_dir}')
